#include "stdutils.h"
#include "complex.h"
#include "matrix.h"
#include "inout.h"
#include "func.h"

#define EXTENT 1
#define HMASK (2 * EXTENT + 1)
#define WMASK HMASK
#define MASKSIZE (HMASK * WMASK)

#define DVARMAX (5.0e-1 * EXTENT * EXTENT)/*double maximum variance*/
#define HALFDVARMAX (5.0e-1 * DVARMAX)/*half double maximum variance*/

#define OBJECT 1
#define BACKGROUND (-1)
#define DONTCARE 0

/*
I = input image.
N = HMASK by WMASK neighborhood.
M = marker image.
Q = output image.
(+) = morphological dilation.
(-) = morphological erosion.
(c) = morphological closing.
(o) = morphological opening.
(t) = morphological top-hat reconstruction.
(b) = morphological bottom-hat reconstruction.
*/

#define NOOP 0/*no change: Q = I*/
#define DILATION 1/*dilation: Q = I (+) N; Q(i,j) = MAX_{k,l} I(i-k,j-l)*/
#define EROSION 2/*erosion: Q = I (-) N; Q(i,j) = MIN_{k,l} I(i-k,j-l)*/
#define CLOSING 3/*closing: Q = [I (+) N] (-) N*/
#define OPENING 4/*opening: Q = [I (-) N] (+) N*/
#define CLOSEOPEN 5/*closing followed by opening*/
#define OPENCLOSE 6/*opening followed by closing*/
#define TOPHAT 7/*top-hat reconstruction: Q(i,j) = MIN{[I (+) N](i,j),I(i,j)*/
#define BOTHAT 8/*bot-hat reconstruction: Q(i,j) = MAX{[I (-) N](i,j),I(i,j)*/
#define DIFFERENCE 9/*difference between objects (for binary images only)*/
#define THINBLACK 10/*thinning black objects (for binary images only)*/
#define THINWHITE 11/*thinning white objects (for binary images only)*/
#define THINBOTH 12/*thinning black and white objects (for binary images only)*/
#define NEGATIVE 13/*negative of input image*/

#define NUMELEMENTS 6
#define INCREMENT (M_PI / 4.0)
#define THRESHOLD 1.0e-2
#define ZERO 0
#define FORTYFIVE 1
#define NINETY 2
#define ONETHIRTYFIVE 3
#define ISOTROPIC 4
#define FULL 5

void main();
void BuildElements();
void Operation();
void NoOp();
void Dilation();
void Erosion();
void Closing();
void Opening();
void CloseOpen();
void OpenClose();
void TopHat();
void BotHat();
void Difference();
void ThinBlack();
void HitMiss();
void Negative();

void
main(argc, argv)
int argc;
char **argv;

{
    char *input_file_name = (char *) AllocateString(256);
    char *output_file_name = (char *) AllocateString(256);
    char *original_file_name = (char *) AllocateString(256);
    int input_file_descriptor;
    int output_file_descriptor;
    int original_file_descriptor;
    int height;
    int width;
    int num_points;
    double *input_data;
    double **input_data_matrix;
    double *output_data;
    double **output_data_matrix;
    double *original_data;
    double **original_data_matrix;
    double ***elements;
    int index;
    int row;
    int col;
    int zoom;
    int operation;
    int orientation;
    double mean;

    if(argc < 7) {
	(void) fprintf(stdout, 
	"Syntax: morphology infile outfile height width zoom operation ");
	(void) fprintf(stdout, 
	"orientation\n");
	(void) fprintf(stdout, 
	"Choices of operation:\n");
	(void) fprintf(stdout,
	"   %d -> NO-OP;\n",
	NOOP);
	(void) fprintf(stdout,
	"   %d -> dilation;\n",
	DILATION);
	(void) fprintf(stdout,
	"   %d -> erosion;\n",
	EROSION);
	(void) fprintf(stdout, 
	"   %d -> closing;\n",
	CLOSING);
	(void) fprintf(stdout, 
	"   %d -> opening;\n",
	OPENING);
	(void) fprintf(stdout, 
	"   %d -> closing followed by opening;\n",
	CLOSEOPEN);
	(void) fprintf(stdout, 
	"   %d -> opening followed by closing;\n",
	OPENCLOSE);
	(void) fprintf(stdout, 
	"   %d -> top-hat reconstruction;\n",
	TOPHAT);
	(void) fprintf(stdout, 
	"   %d -> bottom-hat reconstruction;\n",
	BOTHAT);
	(void) fprintf(stdout, 
	"   %d -> set difference (for binary images only);\n",
	DIFFERENCE);
	(void) fprintf(stdout, 
	"   %d -> thinning black objects (for binary images only);\n",
	THINBLACK);
	(void) fprintf(stdout, 
	"   %d -> negative of input image;\n",
	NEGATIVE);
	(void) fprintf(stdout,
	"Default is %d.\n",
	NOOP);
	(void) fprintf(stdout, 
	"Choices of orientation:\n");
	(void) fprintf(stdout,
	"   %d -> 0 degrees;\n",
	ZERO);
	(void) fprintf(stdout,
	"   %d -> 45 degrees;\n",
	FORTYFIVE);
	(void) fprintf(stdout,
	"   %d -> 90 degrees;\n",
	NINETY);
	(void) fprintf(stdout,
	"   %d -> 135 degrees;\n",
	ONETHIRTYFIVE);
	(void) fprintf(stdout,
	"   %d -> isotropic;\n",
	ISOTROPIC);
	(void) fprintf(stdout,
	"   %d -> full mask;\n",
	FULL);
	(void) fprintf(stdout,
	"Default is %d.\n",
	ISOTROPIC);
	(void) fflush(stdout);
	(void) exit(0);
    }
    (void) sscanf(argv[1], "%s", input_file_name);
    (void) sscanf(argv[2], "%s", output_file_name);
    (void) sscanf(argv[3], "%d", &height);
    (void) sscanf(argv[4], "%d", &width);
    (void) sscanf(argv[5], "%d", &zoom);
    if(argc > 6) {
	(void) sscanf(argv[6], "%d", &operation);
    } else {
	operation = NOOP;
    }
    if(argc > 7) {
	(void) sscanf(argv[7], "%d", &orientation);
	if(!((orientation == ISOTROPIC) || (orientation == FULL) ||
	(orientation == ZERO) || (orientation == FORTYFIVE) ||
	(orientation == NINETY) || (orientation == ONETHIRTYFIVE))) {
	    (void) fprintf(stdout,
	    "Error: Illegal orientation %d.  Exit 1.\n",
	    orientation);
	    (void) fflush(stdout);
	    (void) exit(1);
	}
    } else {
	orientation = ISOTROPIC;
    }
    if(!((operation == NOOP) ||
    (operation == DILATION) || (operation == EROSION) ||
    (operation == CLOSING) || (operation == OPENING) ||
    (operation == CLOSEOPEN) || (operation == OPENCLOSE) ||
    (operation == TOPHAT) || (operation == BOTHAT) ||
    (operation == DIFFERENCE) || (operation == THINBLACK) ||
    (operation == NEGATIVE))) {
	(void) fprintf(stdout,
	"Error: Illegal morphological operation %d.  Exit 1.\n",
	operation);
	(void) fflush(stdout);
	(void) exit(1);
    }
    if((operation == TOPHAT) || (operation == BOTHAT) ||
    (operation == DIFFERENCE)) {
	if(argc < 9) {
	    (void) fprintf(stdout, 
	    "Syntax: morphology infile outfile height width zoom operation ");
	    (void) fprintf(stdout,
	    "orientation original\n");
	    (void) fflush(stdout);
	    (void) exit(0);
	}
	(void) sscanf(argv[8], "%s", original_file_name);
	original_file_descriptor = (int) open(original_file_name, O_RDONLY, 0);
	if((num_points = (int) lseek(original_file_descriptor, 0, SEEK_END)) ==
	(int) NIL) {
	    (void) fprintf(stdout, "Error: File %s nonexistent.  Exit 1.\n",
	    original_file_name);
	    (void) fflush(stdout);
	    (void) exit(1);
	}
	(void) lseek(original_file_descriptor, -(num_points), SEEK_END);
	original_data = (double *) RealCreateVector(num_points);
	original_data_matrix = (double **) RealCreateMatrix(height, width);
	(void) ReadIn(original_file_descriptor,
	original_data, original_data_matrix, NOVERTICALNOHORIZONTALFLIP,
	height, width, num_points);
	/**/
	(void) RealScaleMatrix(original_data_matrix, original_data_matrix,
	2.0 / (MAXGRAYVALUE - MINGRAYVALUE), height, width);
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		original_data_matrix[row][col] -= 1.0;
	    }
	}
	/**/
    }
    input_file_descriptor = (int) open(input_file_name, O_RDONLY, 0);
    output_file_descriptor = (int) creat(output_file_name, 0644);
    if((num_points = (int) lseek(input_file_descriptor, 0, SEEK_END)) ==
    (int) NIL) {
	(void) fprintf(stdout, "Error: File %s nonexistent.  Exit 1.\n",
	input_file_name);
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) lseek(input_file_descriptor, -(num_points), SEEK_END);
    input_data = (double *) RealCreateVector(num_points);
    input_data_matrix = (double **) RealCreateMatrix(height, width);
    output_data = (double *) RealCreateVector(num_points);
    output_data_matrix = (double **) RealCreateMatrix(height, width);
    (void) ReadIn(input_file_descriptor, input_data, input_data_matrix,
    NOVERTICALNOHORIZONTALFLIP, height, width, num_points);
    /**/
    (void) RealScaleMatrix(input_data_matrix, 
    input_data_matrix, 2.0 / (MAXGRAYVALUE - MINGRAYVALUE), height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    input_data_matrix[row][col] -= 1.0;
	}
    }
    /**/
    (void) RealAccumulateMatrix(&mean, input_data_matrix, height, width);
    mean /= num_points;
printf("mean=%g\n",mean);fflush(stdout);
    elements = (double ***) SafeMalloc((unsigned) NUMELEMENTS *
    sizeof(double **), "elements", "main");
    for(index = 0; index < NUMELEMENTS; index++) {
	elements[index] = (double **) RealCreateMatrix(HMASK, WMASK);
    }
    (void) BuildElements(elements);
    (void) Operation(output_data_matrix, input_data_matrix,
    original_data_matrix, operation, elements, orientation, height, width);
    (void) RealMatrixToRealVector(output_data, 
    output_data_matrix, height, width);
    (void) WriteOut(output_file_descriptor, NOVERTICALNOHORIZONTALFLIP, zoom, 
    output_data, height, width, num_points, COEFF);
    if((int) close(input_file_descriptor) == (int) NIL) {
	(void) fprintf(stdout, 
	"Error: Cannot close input file '%s'.  Exit 1.\n", input_file_name);
	(void) fflush(stdout);
	(void) exit(1);
    }
    if((int) close(output_file_descriptor) == (int) NIL) {
	(void) fprintf(stdout, 
	"Error: Cannot close output file '%s'.  Exit 1.\n",
	output_file_name);
	(void) fflush(stdout);
	(void) exit(1);
    }
    if((operation == TOPHAT) || (operation == BOTHAT)) {
	if((int) close(original_file_descriptor) == (int) NIL) {
	    (void) fprintf(stdout, 
	    "Error: Cannot close input file '%s'.  Exit 1.\n",
	    original_file_name);
	    (void) fflush(stdout);
	    (void) exit(1);
	}
	(void) RealFreeVector(original_data);
	(void) RealFreeMatrix(original_data_matrix, height);
    }
    (void) RealFreeVector(input_data);
    (void) RealFreeMatrix(input_data_matrix, height);
    (void) RealFreeVector(output_data);
    (void) RealFreeMatrix(output_data_matrix, height);
    for(index = 0; index < NUMELEMENTS; index++) {
	(void) RealFreeMatrix(elements[index], HMASK);
    }
    (void) free(elements);
    (void) free(input_file_name);
    (void) free(output_file_name);
    (void) exit(0);
}

void
BuildElements(elements)
double ***elements;

{
    double theta;
    double c;
    double s;
    double mag;
    double vary;
    double varx;
    double exponent;
    int index;
    int n2;
    int n1;

    for(index = 0; index < NUMELEMENTS - 1; index++) {
	theta = INCREMENT * index;
	c = (double) cos(theta);
	s = (double) sin(theta);
	if(index == ISOTROPIC) {
	    mag = 0.0;
	} else {
	    mag = 5.0e-1;
	}
	vary = HALFDVARMAX * (1.0 - mag);
	varx = DVARMAX - vary;
	for(n2 = -EXTENT; n2 <= EXTENT; n2++) {
	    for(n1 = -EXTENT; n1 <= EXTENT; n1++) {
		exponent = 
		((s * n1 + c * n2) * (s * n1 + c * n2) /
		(vary + EPSILON)) +
		((c * n1 - s * n2) * (c * n1 - s * n2) /
		(varx + EPSILON));
		elements[index][(n2+HMASK)%HMASK][(n1+WMASK)%WMASK] =
		((double) exp(-exponent) > THRESHOLD)
		? 1.0 * OBJECT : 1.0 * BACKGROUND;
	    }
	}
    }
    for(n2 = -EXTENT; n2 <= EXTENT; n2++) {
	for(n1 = -EXTENT; n1 <= EXTENT; n1++) {
	    elements[FULL][(n2+HMASK)%HMASK][(n1+WMASK)%WMASK] = 1.0 * OBJECT;
	}
    }
}

void
Operation(output_data_matrix, input_data_matrix, original_data_matrix,
operation, elements, orientation, height, width)
double **output_data_matrix;
double **input_data_matrix;
double **original_data_matrix;
int operation;
double ***elements;
int orientation;
int height;
int width;

{
    double **output = output_data_matrix;
    double **input = input_data_matrix;
    double **original = original_data_matrix;

    switch(operation) {
	case NOOP:
	    (void) NoOp(output, input, height, width);
	break;
	case DILATION:
	    (void) Dilation(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case EROSION:
	    (void) Erosion(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case CLOSING:
	    (void) Closing(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case OPENING:
	    (void) Opening(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case CLOSEOPEN:
	    (void) CloseOpen(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case OPENCLOSE:
	    (void) OpenClose(output,
	    input, EXTENT, elements, orientation, height, width);
	break;
	case TOPHAT:
	    (void) TopHat(output, input, original, height, width);
	break;
	case BOTHAT:
	    (void) BotHat(output, input, original, height, width);
	break;
	case DIFFERENCE:
	    (void) Difference(output, input, original, height, width);
	break;
	case THINBLACK:
	    (void) ThinBlack(output, input, height, width);
	break;
	case NEGATIVE:
	    (void) Negative(output, input, height, width);
	break;
	default:
	break;
    }
}

void
NoOp(output, input, height, width)
double **output;
double **input;
int height;
int width;

{
    (void) RealCopyMatrix(output, input, height, width);
}

void
Dilation(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **element;
    double **img;
    int row;
    int col;
    int n2;
    int n1;
    double max_gray;

    element = elements[orientation];
    img = (double **) RealCreateMatrix(height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    max_gray = -INFINITY;
	    for(n2 = -extent; n2 <= extent; n2++) {
		for(n1 = -extent; n1 <= extent; n1++) {
		    if((int)
		    FixFloat(element[(n2+HMASK)%HMASK][(n1+WMASK)%WMASK]) == 
		    OBJECT) {
			max_gray = (double) MaxFloat(max_gray,
			input[(row-n2+height)%height][(col-n1+width)%width]);
		    }
		}
	    }
	    img[row][col] = max_gray;
	}
    }
    (void) RealCopyMatrix(output, img, height, width);
    (void) RealFreeMatrix(img, height);
}

void
Erosion(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **element;
    double **img;
    int row;
    int col;
    int n2;
    int n1;
    double min_gray;

    element = elements[orientation];
    img = (double **) RealCreateMatrix(height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    min_gray = INFINITY;
	    for(n2 = -extent; n2 <= extent; n2++) {
		for(n1 = -extent; n1 <= extent; n1++) {
		    if((int)
		    FixFloat(element[(n2+HMASK)%HMASK][(n1+WMASK)%WMASK]) == 
		    OBJECT) {
			min_gray = (double) MinFloat(min_gray,
			input[(row-n2+height)%height][(col-n1+width)%width]);
		    }
		}
	    }
	    img[row][col] = min_gray;
	}
    }
    (void) RealCopyMatrix(output, img, height, width);
    (void) RealFreeMatrix(img, height);
}

void
Closing(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **img;

    img = (double **) RealCreateMatrix(height, width);
    (void) Dilation(img, input, extent, elements, orientation, height, width);
    (void) Erosion(output, img, extent, elements, orientation, height, width);
    (void) RealFreeMatrix(img, height);
}

void
Opening(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **img;

    img = (double **) RealCreateMatrix(height, width);
    (void) Erosion(img, input, extent, elements, orientation, height, width);
    (void) Dilation(output, img, extent, elements, orientation, height, width);
    (void) RealFreeMatrix(img, height);
}

void
CloseOpen(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **img;

    img = (double **) RealCreateMatrix(height, width);
    (void) Closing(img, input, extent, elements, orientation, height, width);
    (void) Opening(output, img, extent, elements, orientation, height, width);
    (void) RealFreeMatrix(img, height);
}

void
OpenClose(output, input, extent, elements, orientation, height, width)
double **output;
double **input;
int extent;
double ***elements;
int orientation;
int height;
int width;

{
    double **img;

    img = (double **) RealCreateMatrix(height, width);
    (void) Opening(img, input, extent, elements, orientation, height, width);
    (void) Closing(output, img, extent, elements, orientation, height, width);
    (void) RealFreeMatrix(img, height);
}

void
TopHat(output, input, original, height, width)
double **output;
double **input;
double **original;
int height;
int width;

{
    double ***elements;
    double **img1;
    double **img2;
    int row;
    int col;
    int n2;
    int n1;
    int extent;
    int hmask;
    int wmask;
    int index;
    int count;
    int idempotent;

    extent = 1;
    hmask = (2 * extent + 1);
    wmask = hmask;
    elements = (double ***) SafeMalloc((unsigned) NUMELEMENTS *
    sizeof(double **), "elements", "BotHat");
    for(index = 0; index < NUMELEMENTS; index++) {
	elements[index] = (double **) RealCreateMatrix(hmask, wmask);
    }
    img1 = (double **) RealCreateMatrix(height, width);
    img2 = (double **) RealCreateMatrix(height, width);
    for(n2 = -extent; n2 <= extent; n2++) {
	for(n1 = -extent; n1 <= extent; n1++) {
	    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
	}
    }
    (void) RealCopyMatrix(img1, input, height, width);
    for(count = 0, idempotent = 0; !idempotent; count++) {
	(void) fprintf(stdout, "count=%d\n", count);
	(void) fflush(stdout);
	(void) Dilation(img2, img1, extent, elements, FULL, height, width);
	idempotent = 1;
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		img2[row][col] =
		(double) MinFloat(img2[row][col], original[row][col]);
		idempotent = idempotent && (img2[row][col] == img1[row][col]);
	    }
	}
	(void) RealCopyMatrix(img1, img2, height, width);
    }
    (void) RealCopyMatrix(output, img2, height, width);
    for(index = 0; index < NUMELEMENTS; index++) {
	(void) RealFreeMatrix(elements[index], hmask);
    }
    (void) free(elements);
    (void) RealFreeMatrix(img1, height);
    (void) RealFreeMatrix(img2, height);
}

void
BotHat(output, input, original, height, width)
double **output;
double **input;
double **original;
int height;
int width;

{
    double ***elements;
    double **img1;
    double **img2;
    int row;
    int col;
    int n2;
    int n1;
    int extent;
    int hmask;
    int wmask;
    int index;
    int count;
    int idempotent;

    extent = 1;
    hmask = (2 * extent + 1);
    wmask = hmask;
    elements = (double ***) SafeMalloc((unsigned) NUMELEMENTS *
    sizeof(double **), "elements", "BotHat");
    for(index = 0; index < NUMELEMENTS; index++) {
	elements[index] = (double **) RealCreateMatrix(hmask, wmask);
    }
    img1 = (double **) RealCreateMatrix(height, width);
    img2 = (double **) RealCreateMatrix(height, width);
    for(n2 = -extent; n2 <= extent; n2++) {
	for(n1 = -extent; n1 <= extent; n1++) {
	    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
	}
    }
    (void) RealCopyMatrix(img1, input, height, width);
    for(count = 0, idempotent = 0; !idempotent; count++) {
	(void) fprintf(stdout, "count=%d\n", count);
	(void) fflush(stdout);
	(void) Erosion(img2, img1, extent, elements, FULL, height, width);
	idempotent = 1;
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		img2[row][col] =
		(double) MaxFloat(img2[row][col], original[row][col]);
		idempotent = idempotent && (img2[row][col] == img1[row][col]);
	    }
	}
	(void) RealCopyMatrix(img1, img2, height, width);
    }
    (void) RealCopyMatrix(output, img2, height, width);
    for(index = 0; index < NUMELEMENTS; index++) {
	(void) RealFreeMatrix(elements[index], hmask);
    }
    (void) free(elements);
    (void) RealFreeMatrix(img1, height);
    (void) RealFreeMatrix(img2, height);
}

void
Difference(output, input1, input2, height, width)
double **output;
double **input1;
double **input2;
int height;
int width;

{
    int row;
    int col;

    (void) RealSubtractMatrices(output, input1, input2, height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    if((double) AbsFloat(output[row][col]) > EPSILON) {
		output[row][col] = 1.0 * OBJECT;
	    } else {
		output[row][col] = 1.0 * BACKGROUND;
	    }
	}
    }
}

void
ThinBlack(output, input, height, width)
double **output;
double **input;
int height;
int width;

{
    double ***elements;
    double **img1;
    int n2;
    int n1;
    int extent;
    int hmask;
    int wmask;
    int index;

    extent = 1;
    hmask = (2 * extent + 1);
    wmask = hmask;
    elements = (double ***) SafeMalloc((unsigned) NUMELEMENTS *
    sizeof(double **), "elements", "BotHat");
    for(index = 0; index < NUMELEMENTS; index++) {
	elements[index] = (double **) RealCreateMatrix(hmask, wmask);
    }
    img1 = (double **) RealCreateMatrix(height, width);
    n2 = -1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n2 = 0;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n2 = 1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    (void) HitMiss(img1, input, elements, FULL, height, width);
    (void) Difference(output, input, img1, height, width);
/*
RealCopyMatrix(output,img1,height,width);
return;
*/
    n2 = -1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n2 = 0;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n2 = 1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    (void) HitMiss(img1, output, elements, FULL, height, width);
    (void) Difference(output, output, img1, height, width);
    n2 = -1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n2 = 0;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n2 = 1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    (void) HitMiss(img1, output, elements, FULL, height, width);
    (void) Difference(output, output, img1, height, width);
    n2 = -1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n2 = 0;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    n2 = 1;
    n1 = -1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * BACKGROUND;
    n1 = 0;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * DONTCARE;
    n1 = 1;
    elements[FULL][(n2+hmask)%hmask][(n1+wmask)%wmask] = 1.0 * OBJECT;
    (void) HitMiss(img1, output, elements, FULL, height, width);
    (void) Difference(output, output, img1, height, width);
    for(index = 0; index < NUMELEMENTS; index++) {
	(void) RealFreeMatrix(elements[index], hmask);
    }
    (void) free(elements);
    (void) RealFreeMatrix(img1, height);
}

void
HitMiss(output, input, elements, orientation, height, width)
double **output;
double **input;
double ***elements;
int orientation;
int height;
int width;

{
    double **img1;
    double **img2;
    int extent;
    int hmask;
    int wmask;

    extent = 1;
    hmask = (2 * extent + 1);
    wmask = hmask;
    img1 = (double **) RealCreateMatrix(height, width);
    img2 = (double **) RealCreateMatrix(height, width);
    (void) Erosion(img1, input, extent, elements, orientation, height, width);
    (void) Negative(elements[FULL], elements[FULL], hmask, wmask);
    (void) Negative(img2, input, height, width);
    (void) Dilation(img2, img2, extent, elements, orientation, height, width);
    (void) Negative(img2, img2, height, width);
    (void) Difference(output, img1, img2, height, width);
    (void) RealFreeMatrix(img1, height);
    (void) RealFreeMatrix(img2, height);
}

void
Negative(out, in, n, m)
double **out;
double **in;
int n;
int m;

{
    (void) RealScaleMatrix(out, in, (-1.0), n, m);
}
