
/*******************************************************************************************/
/*                        C ROUTINES FOR THE STEERABLE PYRMAID
                   COPYRIGHT 1993, PERCEPTUAL COMPUTING GROUP, MEDIA LABORATORY
		           MASSACHUSSETTS INSTITUTE OF TECHNOLOGY

*/


#include <stdio.h>
#include <math.h>

void twod_conv_s( );
void twod_conv_s_circ();
void twod_conv_s_refl();
void twod_conv_ns();
void twod_conv_ns_circ();
void twod_conv_ns_refl();
void oned_conv();


/*-----------------------------------------------------------*/
/*                2D separable convolution                   */
/*-----------------------------------------------------------*/

/* assume user passes tmp of sizeof(im1) */

void twod_conv_s(im1,res,tmp,xfilt,yfilt,xdim,ydim,xfdim,yfdim,edge,add)
     float **im1;     /* input image */
     float **res;     /* output image */
     float **tmp;     /* temp image */
     float *xfilt, *yfilt;  /* separable filters */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int edge;  /* edge handling method - either circ or refl */
     int add ;  /* whether to add resulting image to res */
{
     if (xfdim == 0 && yfdim == 0) {
       printf("Can't do the convolution \n");
       exit(1);
     }
     
     if (edge == 1) { /* circular convolution */
       twod_conv_s_circ(im1,res,tmp,xfilt,yfilt,xdim,ydim,xfdim,yfdim,add);
       }
     else if (edge == 0) { /* handle edges by reflecting about edge pixel */
       twod_conv_s_refl(im1,res,tmp,xfilt,yfilt,xdim,ydim,xfdim,yfdim,add);}
    
     else  { 
       fprintf(stderr, "unknown edge method %d.\n",edge);
       exit(-1);
     }
   }


/*---------------------------------------------------------------*/
/*               2D separable convolution with circular edge     */
/*---------------------------------------------------------------*/

void twod_conv_s_circ(im1,res,tmp,xfilt,yfilt,xdim,ydim,xfdim,yfdim,add)
     float **im1;     /* input image */
     float **res;     /* output image */
     float **tmp;     /* temp image */
     float *xfilt, *yfilt;  /* separable filters */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int add ;  /* whether to add resulting image to res */
{

  register int i,j;
  register int x_off, y_off;
  register int x_pos, y_pos;

  register int xhalf = xfdim/2;
  register int xhigh = xfdim - xhalf;
  register int yhalf = yfdim/2;
  register int yhigh = yfdim - yhalf;

  register float   sum;


  /* convolve in x */
  
  for (i=0;i<ydim;i++) {
    for (j=0; j<xdim; j++) {
      sum = 0.0;
      
      for (x_off = -xhalf; x_off<xhigh; x_off++) {
	x_pos = j+x_off;
	if (x_pos < 0) x_pos += xdim;
	else if (x_pos >= xdim) x_pos -= xdim;
	sum += im1[i][x_pos]*xfilt[x_off+xhalf];
      }

      tmp[i][j]=sum;
    }
  }
  
/* convolve in y */
  
  for (i=0;i<ydim; i++) {
    for (j=0; j<xdim; j++) {
      sum=0.0;

      for (y_off = -yhalf; y_off<yhigh; y_off++) {
	y_pos = i+y_off;
        if (y_pos < 0) y_pos += ydim;
        else if (y_pos >= ydim) y_pos -= ydim;
	sum += tmp[y_pos][j]*yfilt[y_off+yhalf];
      }
      if (add)
	res[i][j]+=sum;
      else
	res[i][j] = sum;
    }
  }
}



/* -------------------------------------------------------------------------*/
/*    Separable convolution with reflection about pixel edge                */
/* -------------------------------------------------------------------------*/


void twod_conv_s_refl(im1,res,tmp,xfilt,yfilt,xdim,ydim,xfdim,yfdim,add)
     float **im1;     /* input image */
     float **res;     /* output image */
     float **tmp;     /* temp image */
     float *xfilt, *yfilt;  /* separable filters */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int add ;  /* whether to add resulting image to res */
{

  register int i,j,k1,k2;
  register int x_off, y_off;
  register int x_pos, y_pos;

  register int xhalf = xfdim/2;
  register int xhigh = xfdim - xhalf;
  register int yhalf = yfdim/2;
  register int yhigh = yfdim - yhalf;
 
  register float   sum;


  /* convolve in x */
  
  for (i=0;i<ydim;i++) {
    for (j=0; j<xdim; j++) {
      sum = 0.0;
     
      k1=k2=1;
      for (x_off = -xhalf; x_off<xhigh; x_off++) {
	x_pos = j+x_off;
	if (x_pos < 0) {
	  x_pos = j+k1;
	  k1++;
	}

	else if (x_pos >= xdim) {
	  x_pos = j- k2;
          k2++;
	}
	sum += im1[i][x_pos]*xfilt[x_off+xhalf];
      }
      tmp[i][j]=sum;
    }
  }
  
/* convolve in y */
  
  for (i=0;i<ydim; i++) {
    for (j=0; j<xdim; j++) {
      sum=0.0;

      k1=k2=1;
      for (y_off = -yhalf; y_off<yhigh; y_off++) {
	y_pos = i+y_off;
        if (y_pos < 0) {
	  y_pos=i+k1;
          k1++;
        }
        else if (y_pos >= ydim) {
	  y_pos = i - k2;
          k2++;
	}
	sum += tmp[y_pos][j]*yfilt[y_off+yhalf];
      }
      
      if (add)
	res[i][j] +=sum;
      else
	res[i][j] = sum;
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*                       2D Non-separable Convolution                          */  
/*-----------------------------------------------------------------------------*/
void twod_conv_ns(im1,res,tmp,filt,xdim,ydim,xfdim,yfdim,edge,add)
     
     float *im1;     /* input image */
     float *res;     /* output image */
     float *tmp;     /* temp image */
     float *filt;  /* separable filters */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int edge;  /* edge handling method - either circ or refl */
     int add ;  /* whether to add resulting image to res */
{
  if (xfdim == 0 && yfdim == 0) {
    printf("Can't do the convolution \n");
    exit(1);
  }
  
  if (edge == 1) {
    /* circular convolution */
   
    twod_conv_ns_circ(im1,res,tmp,filt,xdim,ydim,xfdim,yfdim,add);
  }
  
  if (edge == 0) { /* handle edges by reflecting about edge pixel */
    twod_conv_ns_refl(im1,res,tmp,filt,xdim,ydim,xfdim,yfdim,add);}
  
  if ((edge != 0) && (edge != 1)){ 
    fprintf(stderr, "unknown edge method %d.\n",edge);
    exit(-1);
  }
}


/*--------------------------------------------------------------------------*/
/*                      2D Non-separable convolution with Circular edge     */
/*--------------------------------------------------------------------------*/

void twod_conv_ns_circ(im1,res,tmp,filt,xdim,ydim,xfdim,yfdim,add)
     float **im1;     /* input image */
     float **res;     /* output image */
     float **tmp;     /* temp image */
     float **filt;  /* filter image */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int add ;  /* whether to add resulting image to res */
{
  
  register int i,j;
  register int x_off, y_off;
  register int x_pos, y_pos;

  register int xhalf = xfdim/2;
  register int xhigh = xfdim - xhalf;
  register int yhalf = yfdim/2;
  register int yhigh = yfdim - yhalf;
 
  register float   sum;
  register int filtsize = xfdim*yfdim;

/* convolve in x,y */

 
  for (i=0; i< ydim; i++) {
    for (j=0; j< xdim; j++) {
      sum=0.0;

      for (y_off = -yhalf; y_off<yhigh; y_off++) {
	y_pos = i + y_off;
	if (y_pos<0) y_pos +=ydim;
        else if (y_pos >=ydim) y_pos -= ydim;
	
	for (x_off = -xhalf; x_off<xhigh; x_off++) {
	  x_pos = j + x_off;
	  if (x_pos < 0) x_pos += xdim;
	  else if (x_pos >= xdim) x_pos -= xdim;
	
	  sum +=im1[y_pos][x_pos] * filt[y_off + yhalf][x_off + xhalf];
	 
	}
      }
      if (add)
	res[i][j] += sum;
      else
	res[i][j] =sum;
    }
   
  }
 
}

/*-----------------------------------------------------------------------*/
/*       2D non-separable convolution with reflection about edge pixel   */
/*-----------------------------------------------------------------------*/

void twod_conv_ns_refl(im1,res,tmp,filt,xdim,ydim,xfdim,yfdim,add)
     float **im1;     /* input image */
     float **res;     /* output image */
     float **tmp;     /* temp image */
     float **filt;    /* filter image */
     int xdim, ydim ;  /* image dimensions */
     int xfdim, yfdim;  /* filter dimensions */
     int add ;  /* whether to add resulting image to res */
{
  
  register int i,j,k1,k2,k3,k4;
  register int x_off, y_off;
  register int x_pos, y_pos;

  register int xhalf = xfdim/2;
  register int xhigh = xfdim - xhalf;
  register int yhalf = yfdim/2;
  register int yhigh = yfdim - yhalf;
 
  register float   sum;
  register int filtsize = xfdim;

/* convolve in x,y */

  for (i=0; i< ydim; i++) {
    for (j=0; j< xdim; j++) {
      sum=0.0;
      k1=k2=1;
      for (y_off = -yhalf; y_off<yhigh; y_off++) {
	y_pos = i + y_off;
	if (y_pos<0) {
	  y_pos =i+k1;
	  k1++;
	}
        else if (y_pos >=ydim) {
	  y_pos = i-k2;
	  k2++;
	}

	k3=k4=1;
	for (x_off = -xhalf; x_off<xhigh; x_off++) {
	  x_pos = j + x_off;
	  if (x_pos < 0) {
	    x_pos=j+k3;
            k3++;
	  }
	  else if (x_pos >= xdim) {
	    x_pos=j-k4;
	    k4++;
	  }

	  sum += im1[y_pos][x_pos] * filt[y_off + yhalf][x_off + xhalf];
	}
      }
      if (add)
	res[i][j] += sum;
      else
	res[i][j] =sum;
    }
   
  }
}


/*********************************************************************/
/*                   1-D convolution with circular edge              */
/*********************************************************************/

void oned_conv(signal,output,filt,signal_size,filt_size)
     float signal[]; /* input signal  */
     float output[]; /* output signal */
     float filt[];   /* filter   */
     int signal_size,filt_size;  /* input and filter sizes */
{

  register int i;
  register int x_off;
  register int x_pos;

  register int xhalf = filt_size/2;
  register int xhigh = filt_size - xhalf;
 
  register float sum;

/* convolve  */

for (i=0;i<signal_size;i++)
  {
    sum=0.0;

    for (x_off = -xhalf; x_off<xhigh; x_off++){
      x_pos = i + x_off;
      if (x_pos < 0) x_pos += signal_size;
      else if (x_pos >= signal_size) x_pos -= signal_size;
      sum += signal[x_pos]*filt[x_off+xhalf];
    }
    
    output[i]=sum;
  }

}









