/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "complex.h"
#include "func.h"

/*
This function allocates and fills the COMPLEX ADT. 
*/

COMPLEX *
ComplexBuild(real_part, imaginary_part)
double real_part;
double imaginary_part;

{
    COMPLEX *complex_number;

    complex_number = (COMPLEX *) SafeMalloc(sizeof(COMPLEX), 
    "complex_number", "ComplexBuild");
    complex_number->real_part = real_part;
    complex_number->imaginary_part = imaginary_part;
    return(complex_number);
}

/*
This function frees memory occupied by a COMPLEX ADT. 
*/

void
ComplexFree(a)
COMPLEX *a;

{
    (void) free(a);
}

/*
This function sets a complex number to the given value.
*/

void
ComplexSet(complex_number, real_part, imaginary_part)
COMPLEX *complex_number;
double real_part;
double imaginary_part;

{
    complex_number->real_part = real_part;
    complex_number->imaginary_part = imaginary_part;
}

/*
This function moves the source complex number into the 
destination complex number.
*/

void
ComplexCopy(complex_number_dest, complex_number_src)
COMPLEX *complex_number_dest;
COMPLEX *complex_number_src;

{
    complex_number_dest->real_part = complex_number_src->real_part;
    complex_number_dest->imaginary_part = complex_number_src->imaginary_part;
}

/*
This function swaps two complex numbers.
*/

void
ComplexSwap(complex_number1, complex_number2)
COMPLEX *complex_number1;
COMPLEX *complex_number2;

{
    COMPLEX *complex_temp;

    complex_temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    (void) ComplexCopy(complex_temp, complex_number1);
    (void) ComplexCopy(complex_number1, complex_number2);
    (void) ComplexCopy(complex_number2, complex_temp);
    (void) ComplexFree(complex_temp);
}

/*
This function computes the complex conjugate of a complex number.
*/

void
ComplexConjugate(complex_result, complex_number)
COMPLEX *complex_result;
COMPLEX *complex_number;

{
    double save_real_part;
    double save_imaginary_part;

    save_real_part = complex_number->real_part;
    save_imaginary_part = complex_number->imaginary_part;
    complex_result->real_part = save_real_part;
    complex_result->imaginary_part = (-save_imaginary_part);
}

/*
This function computes the magnitude of a complex number.
*/

double
ComplexMagnitude(complex_number)
COMPLEX *complex_number;

{
    return((double) sqrt(complex_number->real_part * 
    complex_number->real_part + complex_number->imaginary_part * 
    complex_number->imaginary_part));
}

/*
This function scales the complex number.
*/

void
ComplexScale(complex_result, complex_number, scaling_factor)
COMPLEX *complex_result;
COMPLEX *complex_number;
double scaling_factor;

{
    double save_real_part;
    double save_imaginary_part;

    save_real_part = complex_number->real_part;
    save_imaginary_part = complex_number->imaginary_part;
    complex_result->real_part = scaling_factor * save_real_part;
    complex_result->imaginary_part = scaling_factor * save_imaginary_part;
}

/*
This function adds two complex numbers.
*/

void
ComplexAdd(complex_result, complex_number1, complex_number2)
COMPLEX *complex_result;
COMPLEX *complex_number1;
COMPLEX *complex_number2;

{
    double save_real_part1;
    double save_imaginary_part1;
    double save_real_part2;
    double save_imaginary_part2;

    save_real_part1 = complex_number1->real_part;
    save_imaginary_part1 = complex_number1->imaginary_part;
    save_real_part2 = complex_number2->real_part;
    save_imaginary_part2 = complex_number2->imaginary_part;
    complex_result->real_part = 
    save_real_part1 + save_real_part2;
    complex_result->imaginary_part = 
    save_imaginary_part1 + save_imaginary_part2;
}

/*
This function subtracts the second complex number from the first complex 
number.
*/

void
ComplexSubtract(complex_result, complex_number1, complex_number2)
COMPLEX *complex_result;
COMPLEX *complex_number1;
COMPLEX *complex_number2;

{
    double save_real_part1;
    double save_imaginary_part1;
    double save_real_part2;
    double save_imaginary_part2;

    save_real_part1 = complex_number1->real_part;
    save_imaginary_part1 = complex_number1->imaginary_part;
    save_real_part2 = complex_number2->real_part;
    save_imaginary_part2 = complex_number2->imaginary_part;
    complex_result->real_part = 
    save_real_part1 - save_real_part2;
    complex_result->imaginary_part = 
    save_imaginary_part1 - save_imaginary_part2;
}

/*
This function multiplies two complex numbers.
*/

void
ComplexMultiply(complex_result, complex_number1, complex_number2)
COMPLEX *complex_result;
COMPLEX *complex_number1;
COMPLEX *complex_number2;

{
    double save_real_part1;
    double save_imaginary_part1;
    double save_real_part2;
    double save_imaginary_part2;

    save_real_part1 = complex_number1->real_part;
    save_imaginary_part1 = complex_number1->imaginary_part;
    save_real_part2 = complex_number2->real_part;
    save_imaginary_part2 = complex_number2->imaginary_part;
    complex_result->real_part = 
    save_real_part1 * save_real_part2 -
    save_imaginary_part1 * save_imaginary_part2;
    complex_result->imaginary_part = 
    save_imaginary_part1 * save_real_part2 +
    save_real_part1 * save_imaginary_part2;
}

/*
This function divides the first complex number by the second complex number.
*/

void
ComplexDivide(complex_result, complex_number1, complex_number2)
COMPLEX *complex_result;
COMPLEX *complex_number1;
COMPLEX *complex_number2;

{
    COMPLEX *denom_conj;
    double denom_mag;

    denom_conj = (COMPLEX *) ComplexBuild(0.0, 0.0);
    denom_mag = (double) ComplexMagnitude(complex_number2);
    (void) ComplexConjugate(denom_conj, complex_number2);
    (void) ComplexMultiply(complex_result, complex_number1, denom_conj);
    (void) ComplexScale(complex_result, complex_result, 
    (1.0 / ((denom_mag * denom_mag) + EPSILON)));
    (void) ComplexFree(denom_conj);
}
