/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "convolve.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function performs a circular convolution of two 1-D sequences.
*/

void
CircConv1D(dest_sequence, src_sequence1, src_sequence2, n)
double *dest_sequence;
double *src_sequence1;
double *src_sequence2;
int n;

{
    double *temp_sequence;
    double value;
    int index;

    temp_sequence = (double *) RealCreateVector(n);
    (void) CircFlip1D(temp_sequence, src_sequence2, n);
    for(index = 0; index < n; index++) {
	(void) RealInnerProduct(&value, src_sequence1, temp_sequence, n);
	dest_sequence[index] = value;
	(void) CircShift1D(temp_sequence, temp_sequence, n);
    }
    (void) RealFreeVector(temp_sequence);
}

/*
This function performs a circular flip of a 1-D sequence.
*/

void
CircFlip1D(dest_sequence, src_sequence, n)
double *dest_sequence;
double *src_sequence;
int n;

{
    double *save_src;
    int index;

    save_src = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src, src_sequence, n);
    for(index = 0; index < n; index++) {
	dest_sequence[index] = save_src[(n-index)%n];
    }
    (void) RealFreeVector(save_src);
}

/*
This function performs a circular shift of a 1-D sequence by one in the 
positive direction.
*/

void
CircShift1D(dest_sequence, src_sequence, n)
double *dest_sequence;
double *src_sequence;
int n;

{
    double *save_src;
    int index;

    save_src = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src, src_sequence, n);
    for(index = 0; index < n; index++) {
	dest_sequence[index] = save_src[(index-1+n)%n];
    }
    (void) RealFreeVector(save_src);
}

/*
This function performs a circular convolution of two 2-D sequences.
*/

void
CircConv2D(dest_sequence, src_sequence1, src_sequence2, height, width)
double **dest_sequence;
double **src_sequence1;
double **src_sequence2;
int height;
int width;

{
    double **temp_sequence;
    double value;
    int row;
    int col;

    temp_sequence = (double **) RealCreateMatrix(height, width);
    (void) CircFlip2D(temp_sequence, src_sequence2, height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    (void) RealMatrixInnerProduct(&value, 
	    src_sequence1, temp_sequence, height, width);
	    dest_sequence[row][col] = value;
	    (void) CircShiftX2D(temp_sequence, temp_sequence, height, width);
	}
	(void) CircShiftY2D(temp_sequence, temp_sequence, height, width);
    }
    (void) RealFreeMatrix(temp_sequence, height);
}

/*
This function performs a circular flip of a 2-D sequence.
*/

void
CircFlip2D(dest_sequence, src_sequence, height, width)
double **dest_sequence;
double **src_sequence;
int height;
int width;

{
    double **save_src;
    int row;
    int col;

    save_src = (double **) RealCreateMatrix(height, width);
    (void) RealCopyMatrix(save_src, src_sequence, height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    dest_sequence[row][col] = 
	    save_src[(height-row)%height][(width-col)%width];
	}
    }
    (void) RealFreeMatrix(save_src, height);
}

/*
This function performs a circular shift of a 2-D sequence by one in the 
positive Y direction.
*/

void
CircShiftY2D(dest_sequence, src_sequence, height, width)
double **dest_sequence;
double **src_sequence;
int height;
int width;

{
    double **save_src;
    int row;
    int col;

    save_src = (double **) RealCreateMatrix(height, width);
    (void) RealCopyMatrix(save_src, src_sequence, height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    dest_sequence[row][col] = 
	    save_src[(row-1+height)%height][col];
	}
    }
    (void) RealFreeMatrix(save_src, height);
}

/*
This function performs a circular shift of a 2-D sequence by one in the 
positive X direction.
*/

void
CircShiftX2D(dest_sequence, src_sequence, height, width)
double **dest_sequence;
double **src_sequence;
int height;
int width;

{
    double **save_src;
    int row;
    int col;

    save_src = (double **) RealCreateMatrix(height, width);
    (void) RealCopyMatrix(save_src, src_sequence, height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    dest_sequence[row][col] = 
	    save_src[row][(col-1+width)%width];
	}
    }
    (void) RealFreeMatrix(save_src, height);
}
