/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "fft.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function computes the "real" DFT of a 2-D sequence.  
This function can be used for real and symmetric sequences, the DFT of which
is symmetric and real.  The result will be wrong if the original sequence is
either non-real, or non-symmetric, or both.
*/

void
RDFT2D(dest_sequence, src_sequence, height, width)
double **dest_sequence;
double **src_sequence;
int height;
int width;

{
    COMPLEX ***temp;
    int row;
    int col;

    temp = (COMPLEX ***) ComplexCreateMatrix(height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    (void) ComplexSet(temp[row][col], src_sequence[row][col], 0.0);
	}
    }
    for(row = 0; row < height; row++) {
	(void) DFT1D(temp[row], temp[row], width);
    }
    (void) ComplexTranspose(temp, temp, height, width);
    for(col = 0; col < width; col++) {
	(void) DFT1D(temp[col], temp[col], height);
    }
    (void) ComplexTranspose(temp, temp, width, height);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    dest_sequence[row][col] = temp[row][col]->real_part;
	}
    }
    (void) ComplexFreeMatrix(temp, height, width);
}

/*
This function computes the "real" IDFT of a 2-D sequence.  
This function can be used for real and symmetric sequences, the DFT of which
is symmetric and real.  The result will be wrong if the original sequence is
either non-real, or non-symmetric, or both.
*/

void
RIDFT2D(dest_sequence, src_sequence, height, width)
double **dest_sequence;
double **src_sequence;
int height;
int width;

{
    COMPLEX ***temp;
    int row;
    int col;

    temp = (COMPLEX ***) ComplexCreateMatrix(height, width);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    (void) ComplexSet(temp[row][col], src_sequence[row][col], 0.0);
	}
    }
    for(row = 0; row < height; row++) {
	(void) IDFT1D(temp[row], temp[row], width);
    }
    (void) ComplexTranspose(temp, temp, height, width);
    for(col = 0; col < width; col++) {
	(void) IDFT1D(temp[col], temp[col], height);
    }
    (void) ComplexTranspose(temp, temp, width, height);
    for(row = 0; row < height; row++) {
	for(col = 0; col < width; col++) {
	    dest_sequence[row][col] = temp[row][col]->real_part;
	}
    }
    (void) ComplexFreeMatrix(temp, height, width);
}

/*
This function computes the DFT of a 2-D sequence.
*/

void
DFT2D(dest_sequence, src_sequence, height, width)
COMPLEX ***dest_sequence;
COMPLEX ***src_sequence;
int height;
int width;

{
    int row;
    int col;

    for(row = 0; row < height; row++) {
	(void) DFT1D(dest_sequence[row], src_sequence[row], width);
    }
    (void) ComplexTranspose(dest_sequence, dest_sequence, height, width);
    for(col = 0; col < width; col++) {
	(void) DFT1D(dest_sequence[col], dest_sequence[col], height);
    }
    (void) ComplexTranspose(dest_sequence, dest_sequence, width, height);
}

/*
This function computes the IDFT of a 2-D sequence.
*/

void
IDFT2D(dest_sequence, src_sequence, height, width)
COMPLEX ***dest_sequence;
COMPLEX ***src_sequence;
int height;
int width;

{
    int row;
    int col;

    for(row = 0; row < height; row++) {
	(void) IDFT1D(dest_sequence[row], src_sequence[row], width);
    }
    (void) ComplexTranspose(dest_sequence, dest_sequence, height, width);
    for(col = 0; col < width; col++) {
	(void) IDFT1D(dest_sequence[col], dest_sequence[col], height);
    }
    (void) ComplexTranspose(dest_sequence, dest_sequence, width, height);
}

/*
This function computes the "real" DFT of a 1-D sequence.  
This function can be used for real and symmetric sequences, the DFT of which
is symmetric and real.  The result will be wrong if the original sequence is
either non-real, or non-symmetrix, or both.
*/

void
RDFT1D(dest_sequence, src_sequence, n)
double *dest_sequence;
double *src_sequence;
int n;

{
    COMPLEX **temp;
    int index;

    temp = (COMPLEX **) ComplexCreateVector(n);
    for(index = 0; index < n; index++) {
	(void) ComplexSet(temp[index], src_sequence[index], 0.0);
    }
    (void) Fft(temp, 
    (int) ((double) log(1.0 * n + EPSILON) / (double) log(2.0)));
    for(index = 0; index < n; index++) {
	dest_sequence[index] = temp[index]->real_part;
    }
    (void) ComplexFreeVector(temp, n);
}

/*
This function computes the "real" IDFT of a 1-D sequence.
This function can be used for real and symmetric sequences, the IDFT of which
is symmetric and real.  The result will be wrong if the original sequence is
either non-real, or non-symmetrix, or both.
*/

void
RIDFT1D(dest_sequence, src_sequence, n)
double *dest_sequence;
double *src_sequence;
int n;

{
    double scale;
    COMPLEX **temp;
    int index;

    scale = 1.0 / (n + EPSILON);
    temp = (COMPLEX **) ComplexCreateVector(n);
    for(index = 0; index < n; index++) {
	(void) ComplexSet(temp[index], src_sequence[index], 0.0);
    }
    (void) Fft(temp, 
    (int) ((double) log(1.0 * n + EPSILON) / (double) log(2.0)));
    for(index = 0; index < n; index++) {
	dest_sequence[index] = temp[index]->real_part;
    }
    (void) RealScaleVector(dest_sequence, dest_sequence, scale, n);
    (void) ComplexFreeVector(temp, n);
}

/*
This function computes the DFT of a 1-D sequence.
*/

void
DFT1D(dest_sequence, src_sequence, n)
COMPLEX **dest_sequence;
COMPLEX **src_sequence;
int n;

{
    (void) ComplexConjugateVector(dest_sequence, src_sequence, n);
    (void) Fft(dest_sequence, 
    (int) ((double) log(1.0 * n + EPSILON) / (double) log(2.0)));
    (void) ComplexConjugateVector(dest_sequence, dest_sequence, n);
}

/*
This function computes the IDFT of a 1-D sequence.
*/

void
IDFT1D(dest_sequence, src_sequence, n)
COMPLEX **dest_sequence;
COMPLEX **src_sequence;
int n;

{
    COMPLEX *scale;

    scale = (COMPLEX *) ComplexBuild(1.0 / n, 0.0);
    (void) ComplexCopyVector(dest_sequence, src_sequence, n);
    (void) Fft(dest_sequence, 
    (int) ((double) log(1.0 * n + EPSILON) / (double) log(2.0)));
    (void) ComplexScaleVector(dest_sequence, dest_sequence, scale, n);
    (void) ComplexFree(scale);
}

/*
This function performs the Fast Inverse Discrete Fourier Transform:
N * Finverse(N).  N must be a power of 2.
The FFT algorithm computes N * IDFT(X(index)) (synthesis); therefore, when the 
forward (analysis) DFT is desired, the complex conjugate of the input data 
should be entered into the appropriate ADT.  In this case, in order to obatin 
the DFT, the compex conjugate of the output of the FFT algorithm used (IDFT) 
should be computed.
*/

void
Fft(a, m)
COMPLEX **a;
int m;

{
    COMPLEX *u;
    COMPLEX *w;
    COMPLEX *t;
    int nv2;
    int nm1;
    int j;
    int i;
    int k;
    double ang;
    int ip;
    int l;
    int le;
    int le1;
    int n;

    u = (COMPLEX *) ComplexBuild(0.0, 0.0);
    w = (COMPLEX *) ComplexBuild(0.0, 0.0);
    t = (COMPLEX *) ComplexBuild(0.0, 0.0);
    n = 1 << m;
    nv2 = n / 2;
    nm1 = n - 1;
    j = 0;
    for(i = 0; i < nm1; i++) {
	if(i < j) {
	    (void) ComplexSwap(a[i], a[j]);
	} 
	k = nv2;
	for(; (k - 1) < j;) {
	    j = j - k;
	    k = k / 2;
	}
	j = j + k;
    }
    for(l = 0; l < m; l++) {
	le = (int) FixFloat((double) pow(2.0, 1.0 * (l + 1)));
	le1 = le / 2;
	(void) ComplexSet(u, 1.0, 0.0);
	ang = M_PI / le1;
	(void) ComplexSet(w, (double) cos(ang), (double) sin(ang));
	for(j = 0; j < le1; j++) {
	    for(i = j; i < n; i += le) {
		ip = i + le1;
		(void) ComplexMultiply(t, a[ip], u);
		(void) ComplexSubtract(a[ip], a[i], t);
		(void) ComplexAdd(a[i], a[i], t);
	    }
	    (void) ComplexMultiply(u, u, w);
	}
    }
    (void) ComplexFree(u);
    (void) ComplexFree(w);
    (void) ComplexFree(t);
} 
