/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "inout.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function reads in an image file into an array of real numbers.
*/

void
ReadIn(input_file_descriptor, data, data_matrix, flip, 
height, width, num_points)
int input_file_descriptor;
double *data;
double **data_matrix;
int flip;
int height;
int width;
int num_points;

{
    int index;
    unsigned char *input_file_contents;

    input_file_contents = (unsigned char *) 
    SafeMalloc((unsigned) num_points * sizeof(unsigned char),
    "input_file_contents", "ReadIn");
    if((int) read(input_file_descriptor, input_file_contents, num_points) != 
    num_points) {
	(void) fprintf(stdout, "Error: Input file corrupted.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
    if(num_points != height * width) {
	(void) fprintf(stdout, "Error: Bad input file dimensions.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
    for(index = 0; index < num_points; index++) {
	data[index] = 1.0 * input_file_contents[index];
    }
    (void) FlipImage(data, flip, height, width, num_points);
    (void) RealVectorToRealMatrix(data_matrix, data, height, width);
    (void) free(input_file_contents);
}

/*
This function writes out an array of real numbers into an image file.
*/

void
WriteOut(output_file_descriptor, 
flip, zoom, data, height, width, num_points, style)
int output_file_descriptor;
int flip;
int zoom;
double *data;
int height;
int width;
int num_points;
int style;

{
    unsigned char *output_file_contents;

    output_file_contents = (unsigned char *) 
    SafeMalloc((unsigned) num_points * sizeof(unsigned char),
    "output_file_contents", "WriteOut");
    (void) FlipImage(data, flip, height, width, num_points);
    if(style == SIGNAL) {
	(void) FloatToGraySignalVector(output_file_contents, data, num_points);
    } else if(style == COEFF) {
	(void) FloatToGrayCoeffVector(output_file_contents, data, num_points);
    } else if(style == MAGCOEFF) {
	(void) FloatToGrayMCVector(output_file_contents, data, num_points);
    } else if(style == LOGMAGCOEFF) {
	(void) FloatToGrayLMCVector(output_file_contents, data, num_points);
    } else {
	(void) fprintf(stdout, "Error: Unknown sequence style.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) ZoomImage(&output_file_contents, 
    zoom, &height, &width, &num_points);
    if((int) write(output_file_descriptor, output_file_contents, num_points) != 
    num_points) {
	(void) fprintf(stdout, "Error: Output file corrupted.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) free(output_file_contents);
}

/*
This function displays an array of real numbers as an image onto the screen
using the RLE package.  Warning: the RLE package is not universally available.
*/

void
DisplayRLE(pipe, directive, flip, 
data, title, height, width, num_points, style)
FILE **pipe;
int directive;
int flip;
double *data;
char *title;
int height;
int width;
int num_points;
int style;

{
    unsigned char *output_file_contents;
    char *rle_command;
    double zoom;

    output_file_contents = (unsigned char *) 
    SafeMalloc((unsigned) num_points * sizeof(unsigned char),
    "output_file_contents", "DisplayRLE");
    (void) FlipImage(data, flip, height, width, num_points);
    if(style == SIGNAL) {
	(void) FloatToGraySignalVector(output_file_contents, data, num_points);
    } else if(style == COEFF) {
	(void) FloatToGrayCoeffVector(output_file_contents, data, num_points);
    } else if(style == MAGCOEFF) {
	(void) FloatToGrayMCVector(output_file_contents, data, num_points);
    } else if(style == LOGMAGCOEFF) {
	(void) FloatToGrayLMCVector(output_file_contents, data, num_points);
    } else {
	(void) fprintf(stdout, "Error: Unknown sequence style.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
    if((height == width) && (height < TYPIMAGEDIM)) {
	zoom = (1.0 * TYPIMAGEDIM) / (1.0 * height + EPSILON);
    } else {
	zoom = 1.0;
    }
    /*
    A general RLE command (not used here; just listed for reference).
    (void) sprintf(rle_command,
    "rawtorle -n 1 -h %d -w %d | rleflip -v | rlezoom %E | \
    getx11 -g 1.0 -j -m -T -t %s", height, width, zoom, title);
    */
    (void) ZoomImage(&output_file_contents, 
    (int) FixFloat(zoom), &height, &width, &num_points);
    if((directive == OPENPIPE) || (directive == OPENCLOSEPIPE)) {
	rle_command = (char *) AllocateString(121);
	(void) sprintf(rle_command,
	"%s %d %d %E %s",
	DISPLAYSCRIPT, height, width, zoom, title);
	*pipe = (FILE *) nshpopen(rle_command, "w");
	(void) free(rle_command);
    } 
    /*
    Due to bugs in the RLE package, the actual 'DISPLAYSCRIPT' does not
    use the 'rleflip' and the 'rlezoom' utilities; therefore, the zoom factor
    that the above command contains is retained only for the sake of 
    completeness, without actually being used.  The working and reliable
    flip and zoom utilities are provided by this "inout.c" module.
    */
    (void) fwrite(output_file_contents, 
    sizeof(unsigned char), num_points, *pipe);
    (void) fflush(*pipe);
    if((directive == CLOSEPIPE) || (directive == OPENCLOSEPIPE)) {
	(void) nshpclose(*pipe);
    }
    (void) free(output_file_contents);
}

/*
This function flips the image according to the directive. 
The 1-D data contains the concatenation of the rows of the 2-D image.
*/

void
FlipImage(data, flip, height, width, num_points)
double *data;
int flip;
int height;
int width;
int num_points;

{
    double *save_data;
    int row;
    int col;

    save_data = (double *) RealCreateVector(num_points);
    (void) RealCopyVector(save_data, data, num_points);
    if(flip == VERTICALFLIP) {
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		data[width * row + col] = 
		save_data[width * (height - 1 - row) + col];
	    }
	}
    } else if(flip == HORIZONTALFLIP) {
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		data[width * row + col] = 
		save_data[width * row + width - 1 - col];
	    }
	}
    } else if(flip == VERTICALHORIZONTALFLIP) {
	for(row = 0; row < height; row++) {
	    for(col = 0; col < width; col++) {
		data[width * row + col] =
		save_data[width * (height - 1 - row) + width - 1 - col];
	    }
	}
    } else if(flip != NOVERTICALNOHORIZONTALFLIP) {
	(void) fprintf(stdout, 
	"Error: Illegal flip directive %d\n.  Exit 1.\n", flip);
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) RealFreeVector(save_data);
}

/*
This function zooms the image -- identically along both the 'x' and the 'y' 
axes -- by the given zoom factor.  The zoom factor must be a positive integer.
The 1-D gray image contains the concatenation of the rows of the 2-D image.
*/

void
ZoomImage(gray_image, zoom, height, width, num_points)
unsigned char **gray_image;
int zoom;
int *height;
int *width;
int *num_points;

{
    unsigned char *save_gray_image;
    int save_height;
    int save_width;
    int row;
    int col;
    int n2;
    int n1;
    int index;

    if(zoom < 1) {
	(void) fprintf(stdout, 
	"Error: Illegal zoom factor '%d'.  Exit 1.\n", zoom);
	(void) fflush(stdout);
	(void) exit(1);
    }
    save_height = *height;
    save_width = *width;
    save_gray_image = (unsigned char *) 
    SafeMalloc((unsigned) (*num_points) * sizeof(unsigned char),
    "save_gray_image", "ZoomImage");
    for(index = 0; index < (*num_points); index++) {
	save_gray_image[index] = (*gray_image)[index];
    }
    (void) free(*gray_image);
    *height *= zoom;
    *width *= zoom;
    *num_points = (*height) * (*width);
    *gray_image = (unsigned char *) 
    SafeMalloc((unsigned) (*num_points) * sizeof(unsigned char),
    "*gray_image", "ZoomImage");
    for(row = 0; row < save_height; row++) {
	for(col = 0; col < save_width; col++) {
	    for(n2 = 0; n2 < zoom; n2++) {
		for(n1 = 0; n1 < zoom; n1++) {
		    (*gray_image)[zoom*((*width)*row+col)+(*width)*n2+n1] = 
		    save_gray_image[save_width*row+col];
		}
	    }
	}
    }
    (void) free(save_gray_image);
}

/*
This function conditions data for gray level display of processed signal.
*/

void
FloatToGraySignalVector(gray_data, float_data, n)
unsigned char *gray_data;
double *float_data;
int n;

{
    int index;
    int value;

    for(index = 0; index < n; index++) {
	if(float_data[index] < 1.0 * NEGHUGE) {
	    gray_data[index] = (unsigned char) MINGRAYVALUE;
	} else if(float_data[index] > 1.0 * POSHUGE) {
	    gray_data[index] = (unsigned char) MAXGRAYVALUE;
	} else {
	    value = (int) FixFloat(float_data[index]);
	    if(value < MINGRAYVALUE) {
		gray_data[index] = (unsigned char) MINGRAYVALUE;
	    } else if(value > MAXGRAYVALUE) {
		gray_data[index] = (unsigned char) MAXGRAYVALUE;
	    } else {
		gray_data[index] = (unsigned char) value;
	    }
	}
    }
}

/*
This function conditions data for gray level display of transform coefficients.
*/

void
FloatToGrayCoeffVector(gray_data, float_data, n)
unsigned char *gray_data;
double *float_data;
int n;

{
    int index;
    int value;
    double min_value;
    double max_value;
    double value_range;
    double gray_range = 1.0 * (MAXGRAYVALUE - MINGRAYVALUE);
    double *temp_data;

    temp_data = (double *) RealCreateVector(n);
    min_value = INFINITY;
    max_value = -INFINITY;
    for(index = 0; index < n; index++) {
	min_value = (double) MinFloat(min_value, float_data[index]);
	max_value = (double) MaxFloat(max_value, float_data[index]);
    }
    value_range = max_value - min_value;
printf("min=%g,max=%g,range=%g\n",min_value,max_value,value_range);fflush(stdout);
    for(index = 0; index < n; index++) {
	temp_data[index] = 1.0 * MINGRAYVALUE + 
	((float_data[index] - min_value) / (value_range + EPSILON)) * 
	gray_range;
    }
    for(index = 0; index < n; index++) {
	value = (int) FixFloat(temp_data[index]);
	if(value < MINGRAYVALUE) {
	    gray_data[index] = (unsigned char) MINGRAYVALUE;
	} else if(value > MAXGRAYVALUE) {
	    gray_data[index] = (unsigned char) MAXGRAYVALUE;
	} else {
	    gray_data[index] = (unsigned char) value;
	}
    }
    (void) RealFreeVector(temp_data);
}

/*
This function conditions data for gray level display of 
magnitudes of transform coefficients.
*/

void
FloatToGrayMCVector(gray_data, float_data, n)
unsigned char *gray_data;
double *float_data;
int n;

{
    int index;
    int value;
    double min_value;
    double max_value;
    double value_range;
    double gray_range = 1.0 * (MAXGRAYVALUE - MINGRAYVALUE);
    double *temp_data;

    temp_data = (double *) RealCreateVector(n);
    min_value = INFINITY;
    max_value = -INFINITY;
    for(index = 0; index < n; index++) {
	min_value = (double) MinFloat(min_value, 
	(double) AbsFloat(float_data[index]));
	max_value = (double) MaxFloat(max_value, 
	(double) AbsFloat(float_data[index]));
    }
    value_range = max_value - min_value;
    for(index = 0; index < n; index++) {
	temp_data[index] = 1.0 * MINGRAYVALUE + 
	(((double) AbsFloat(float_data[index]) - 
	min_value) / (value_range + EPSILON)) * gray_range;
    }
    for(index = 0; index < n; index++) {
	value = (int) FixFloat(temp_data[index]);
	if(value < MINGRAYVALUE) {
	    gray_data[index] = (unsigned char) MINGRAYVALUE;
	} else if(value > MAXGRAYVALUE) {
	    gray_data[index] = (unsigned char) MAXGRAYVALUE;
	} else {
	    gray_data[index] = (unsigned char) value;
	}
    }
    (void) RealFreeVector(temp_data);
}

/*
This function conditions data for gray level display of 
logarithm base 10 of magnitudes of transform coefficients.
*/

void
FloatToGrayLMCVector(gray_data, float_data, n)
unsigned char *gray_data;
double *float_data;
int n;

{
    int index;
    int value;
    double min_value;
    double max_value;
    double value_range;
    double gray_range = 1.0 * (MAXGRAYVALUE - MINGRAYVALUE);
    double *temp_data;

    temp_data = (double *) RealCreateVector(n);
    min_value = INFINITY;
    max_value = -INFINITY;
    for(index = 0; index < n; index++) {
	min_value = (double) MinFloat(min_value, 
	(double) log10((double) AbsFloat(float_data[index]) + EPSILON));
	max_value = (double) MaxFloat(max_value, 
	(double) log10((double) AbsFloat(float_data[index]) + EPSILON));
    }
    value_range = max_value - min_value;
    for(index = 0; index < n; index++) {
	temp_data[index] = 1.0 * MINGRAYVALUE + 
	(((double) log10((double) AbsFloat(float_data[index]) + EPSILON) - 
	min_value) / (value_range + EPSILON)) * gray_range;
    }
    for(index = 0; index < n; index++) {
	value = (int) FixFloat(temp_data[index]);
	if(value < MINGRAYVALUE) {
	    gray_data[index] = (unsigned char) MINGRAYVALUE;
	} else if(value > MAXGRAYVALUE) {
	    gray_data[index] = (unsigned char) MAXGRAYVALUE;
	} else {
	    gray_data[index] = (unsigned char) value;
	}
    }
    (void) RealFreeVector(temp_data);
}

/* @(#)popen.c  4.5 (Berkeley) 7/6/84
 * Modified from 4.2, 2.9 BSD popen by Stephen Uitti, PUCC, Nov '85
 * Doesn't invoke shell, saving time, becoming more secure.
 * Calls breakargs to break line into command line arguments,
 * delimited by white space, but ignores globing and quoting
 * characters.  For use where you don't need shell meta character
 * expansion or filename globbing.
 */

#ifdef sparc
#include <vfork.h>
#endif

#define tst(a,b)        (*mode == 'r'? (b) : (a))
#define RDR     0
#define WTR     1
#define SPC '\040'                      /* ascii space */

static  int popen_pid[20];
extern char **environ;
char **breakargs();

/* no-shell popen */
FILE *
nshpopen(cmd, mode)
char *cmd;
char *mode;
{
        int p[2];
        int myside, hisside, pid;
        char **elist;                           /* the execv list */

        elist = breakargs(cmd);
        if (pipe(p) < 0)
                return (NULL);
        myside = tst(p[WTR], p[RDR]);
        hisside = tst(p[RDR], p[WTR]);
        if ((pid = vfork()) == 0) {
                /* myside and hisside reverse roles in child */
                close(myside);
                if (hisside != tst(0, 1)) {     /* 0, 1 => stdin, stdout */
                        dup2(hisside, tst(0, 1));
                        close(hisside);
                }
                execve(elist[0], elist, environ);
                perror("execv");
                _exit(1);                       /* bombed execv, child dies */
        }
        free(elist);                            /* discard the broken args */
        if (pid == -1) {
                close(myside);
                close(hisside);
                return (NULL);
        }
        popen_pid[myside] = pid;
        close(hisside);
        return (fdopen(myside, mode));
}

/* close for nshpopen */
int *
nshpclose(ptr)
FILE *ptr;
{
#ifdef  BSD4_2
        int child, pid, status, omask;

        child = popen_pid[fileno(ptr)];
        fclose(ptr);
#define mask(s) (1 << ((s)-1))
        omask = sigblock(mask(SIGINT)|mask(SIGQUIT)|mask(SIGHUP));
        while ((pid = wait(&status)) != child && pid != -1)
                /* empty */;
        (void) sigsetmask(omask);
        return (int *)(pid == -1 ? -1 : status);
#else           /* 2.9 BSD, at least */
        register f, r;
	void (*hstat)(), (*istat)(), (*qstat)();
        int status;

        f = fileno(ptr);
        fclose(ptr);
        istat = signal(SIGINT, SIG_IGN);
        qstat = signal(SIGQUIT, SIG_IGN);
        hstat = signal(SIGHUP, SIG_IGN);
        while((r = wait(&status)) != popen_pid[f] && r != -1)
                ;
        if(r == -1)
                status = -1;
        signal(SIGINT, istat);
        signal(SIGQUIT, qstat);
        signal(SIGHUP, hstat);
        return (int *)(status);
#endif  BSD4_2
}

/* breakargs - break a string into a string vector for execv.
 * Returns NULL if can't malloc space for vector, else vector.
 * Note, when done with the vector, mearly "free" the vector.
 * Written by Stephen Uitti, PUCC, Nov '85 for the new version
 * of "popen" - "nshpopen", that doesn't use a shell.
 * (used here for the as filters, a newer option).
 *
 * breakargs is copyright (C) Purdue University, 1985
 *
 * put in a fix for cmds lines with "string string" in them
 * Mon Aug 25 13:34:27 EST 1986 (ksb)
 *
 * Permission is hereby given for its free reproduction and
 * modification for non-commercial purposes, provided that this
 * notice and all embedded copyright notices be retained.
 * Commercial organisations may give away copies as part of their
 * systems provided that they do so without charge, and that they
 * acknowledge the source of the software.
 */
#define SPC '\040'                      /* ascii space */

char *
mynext(pch)
register char *pch;
{
        register int fQuote;

        for (fQuote = 0; (*pch != '\000' && *pch != SPC && *pch != '\t')||fQuote
; ++pch) {
                if ('\\' == *pch) {
                        continue;
                }
                switch (fQuote) {
                default:
                case 0:
                        if ('"' == *pch) {
                                fQuote = 1;
                        } else if ('\'' == *pch) {
                                fQuote = 2;
                        }
                        break;
                case 1:
                        if ('"' == *pch)
                                fQuote = 0;
                        break;
                case 2:
                        if ('\'' == *pch)
                                fQuote = 0;
                        break;
                }
        }
        return pch;
}

char **
breakargs(cmd)
char *cmd;
{
        register char *p;               /* tmp                          */
        register char **v;              /* vector of commands returned  */
        register unsigned sum;          /* bytes for malloc             */
        register int i;                 /* number of args               */
        register char *s;               /* save old position            */
        char *malloc(), *strcpy();

        p = cmd;
        while (*p == SPC || *p == '\t')
                p++;
        cmd = p;                        /* no leading spaces            */
        sum = sizeof(char *);
        i = 0;
        while (*p != '\0') {            /* space for argv[];            */
                ++i;
                s = p;
                p = mynext(p);
                sum += sizeof(char *) + 1 + (unsigned)(p - s);
                while (*p == SPC || *p == '\t')
                        p++;
        }
        ++i;
        /* vector starts at v, copy of string follows NULL pointer */
        v = (char **)malloc(sum+1);
        if (v == NULL)
                return (char **)NULL;
        p = (char *)v + i * sizeof(char *); /* after NULL pointer */
        i = 0;                          /* word count, vector index */
        while (*cmd != '\0') {
                v[i++] = p;
                s = cmd;
                cmd = mynext(cmd);
                if (*cmd != '\0')
                        *cmd++ = '\0';
                strcpy(p, s);
                p += strlen(p);
                ++p;
                while (*cmd == SPC || *cmd == '\t')
                        ++cmd;
        }
        v[i] = (char *)NULL;
        return v;
}
