/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "lscsvd.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function decomposes the given [N by M] real-valued matrix A into the 
product of three matrices: U [N by L], D [L by L], and V [L by L], where 
L <= M: A = U * D * V'.  Matrices U and V have orthonormal real-valued column 
vectors.  Matrix D is diagonal with positive values on the main diagonal.
Matrix A corresponds to approximating the real-valued [1 by N] data vector X 
using the LLSE method.  The column vectors of the matrix A are the values of 
M basis functions evaluated at N data points.  Thus the objective is to choose 
a subset of all basis functions, L, and the matrices U, D, and V optimally in 
order to guarantee the required approximation error tolerance E.  The Singular
Value Decomposition process on the combinations of the real-valued column 
vectors of A is used for computing the pseudoinverse matrices.  The result is
the least squares combination.
*/

double
LSCSVD(u, d, v, a, p, l, energy, x, e, n, m)
double **u;
double **d;
double **v;
double **a;
int *p;/*parameter-index list*/
int *l;
double *energy;
double *x;/*data to be approximated*/
double e;
int n;
int m;

{
    int row;
    double error;

    (void) RealInnerProduct(energy, x, x, n);
    e *= *energy;
    for(row = 0; row < m; row++) {
	p[row] = (int) NIL;
    }
printf("energy=%g;target-error=%g\n",*energy,e);fflush(stdout);
    for(*l = 0, error = *energy; ((error > e) && (*l < m)); (*l)++) {
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}
	error = 
	(double) FindLSCombination(u, d, v, p, a, x, *energy, n, m, *l + 1);
    }
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}else{
printf("after basis %d; error=%g\n",(*l),error);fflush(stdout);
}
printf("total %d basis vectors used\n",(*l));fflush(stdout);
    return(error);
}

/*
This function finds the best combination (in the LLSE sense) of L out M 
basis vectors, where L <= M.  The real-valued [N by M] matrix A contains
the original M basis vectors. 
*/

double
FindLSCombination(u, d, v, p, a, x, energy, n, m, l)
double **u;
double **d;
double **v;
int *p;
double **a;
double *x;
double energy;
int n;
int m;
int l;

{
    int row;
    int col;
    double **a_xpose;
    double **b;
    double **b_xpose;
    double **u_xpose;
    int **table;
    int size;
    double loss;
    int best_choice;
    double error;

    a_xpose = (double **) RealCreateMatrix(m, n);
    b = (double **) RealCreateMatrix(n, l);
    b_xpose = (double **) RealCreateMatrix(l, n);
    u_xpose = (double **) RealCreateMatrix(l, n);
    (void) RealTranspose(a_xpose, a, n, m);
    best_choice = (int) NIL;
    size = (int) Choose(m, l);
    table = (int **) SafeMalloc((unsigned) size * sizeof(int *),
    "table", "LSCSVD");
    for(row = 0; row < size; row++) {
	table[row] = (int *) SafeMalloc((unsigned) m * sizeof(int),
	"table[row]", "LSCSVD");
    }
    (void) ConstructChoiceTable(table, size, m, l);
    for(error = energy, row = 0; row < size; row++) {
	for(col = 0; col < l; col++) {
	    (void) RealCopyVector(b_xpose[col], a_xpose[table[row][col]], n);
	}
	(void) RealTranspose(b, b_xpose, l, n);
	(void) SVD(u, d, v, b, n, l);
	(void) RealTranspose(u_xpose, u, n, l);
	loss = (double) TotalError(x, u_xpose, l, energy, n);
	if(loss < error) {
	    error = loss;
	    best_choice = row;
	}
    }
    if(best_choice != (int) NIL) {
	for(col = 0; col < l; col++) {
	    p[col] = table[best_choice][col];
	    (void) RealCopyVector(b_xpose[col], 
	    a_xpose[table[best_choice][col]], n);
	}
	(void) RealTranspose(b, b_xpose, l, n);
	(void) SVD(u, d, v, b, n, l);
    }
    for(row = 0; row < size; row++) {
	(void) free(table[row]);
    }
    (void) RealFreeMatrix(a_xpose, m);
    (void) RealFreeMatrix(b, n);
    (void) RealFreeMatrix(b_xpose, l);
    (void) RealFreeMatrix(u_xpose, l);
    return(error);
}

/*
This function computes the squared error in approximating a [1 by N]
real-valued data vector with L real-valued orthonormal [1 by N] basis vectors.
*/

double
TotalError(x, b, l, energy, n)
double *x;
double **b;
int l;
double energy;
int n;

{
    int index;
    double error;

    for(index = 0, error = energy; index < l; index++) {
	error -= (double) OneGain(x, b[index], energy, n);
    }
    return((double) AbsFloat(error));
}

/*
This function computes the squared gain in approximating a [1 by N]
real-valued data vector with a [1 by N] real-valued orthonormal basis vector.
*/

double
OneGain(x, b, energy, n)
double *x;
double *b;
double energy;
int n;

{
    double projection;

    if(energy < EPSILON) {
	return(energy);
    }
    (void) RealInnerProduct(&projection, b, x, n);
    return(projection * projection);
}
