/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function initializes all elements of the given real-valued vector to zero.
*/

void
RealClearVector(v, n)
double *v;
int n;

{
    int col;

    for(col = 0; col < n; col++) {
	v[col] = 0.0;
    }
}

/*
This function allocates space for a new real-valued vector, initializes 
all of its elements to zero, and returns it on output.
*/

double *
RealCreateVector(n)
int n;

{
    double *v;

    v = (double *) SafeMalloc((unsigned) (n) * sizeof(double),
    "v", "RealCreateVector");
    (void) RealClearVector(v, n);
    return(v);
}

/*
This function frees memory occupied bye a real-valued vector.
*/

void
RealFreeVector(v)
double *v;

{
    (void) free(v);
}

/*
This function copies the contents of the source real-valued vector into the 
destination real-valued vector.
*/

void
RealCopyVector(dest_vector, src_vector, n)
double *dest_vector;
double *src_vector;
int n;

{
    int col;
    double *save_src;

    save_src = (double *) RealCreateVector(n);
    for(col = 0; col < n; col++) {
	save_src[col] = src_vector[col];
    }
    for(col = 0; col < n; col++) {
	dest_vector[col] = save_src[col];
    }
    (void) RealFreeVector(save_src);
}

/*
This function scales the contents of the source real-valued vector into the 
destination real-valued vector.
*/

void
RealScaleVector(dest_vector, src_vector, scale, n)
double *dest_vector;
double *src_vector;
double scale;
int n;

{
    int col;
    double *save_src;

    save_src = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src, src_vector, n);
    for(col = 0; col < n; col++) {
	dest_vector[col] = save_src[col] * scale;
    }
    (void) RealFreeVector(save_src);
}

/*
This function initializes the contents of the destination real-valued vector to
the real-valued scalar constant.
*/

void
RealInitVector(dest_vector, value, n)
double *dest_vector;
double value;
int n;

{
    int col;

    for(col = 0; col < n; col++) {
	dest_vector[col] = value;
    }
}

/*
This function adds the contents of the two source real-valued vectors into 
the destination real-valued vector.
*/

void
RealAddVectors(dest_vector, src_vector1, src_vector2, n)
double *dest_vector;
double *src_vector1;
double *src_vector2;
int n;

{
    int col;
    double *save_src1;
    double *save_src2;

    save_src1 = (double *) RealCreateVector(n);
    save_src2 = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src1, src_vector1, n);
    (void) RealCopyVector(save_src2, src_vector2, n);
    for(col = 0; col < n; col++) {
	dest_vector[col] = save_src1[col] + save_src2[col];
    }
    (void) RealFreeVector(save_src1);
    (void) RealFreeVector(save_src2);
}

/*
This function subtracts the contents of the second source real-valued vector 
from the contents of the first source real-valued vector into the destination 
real-valued vector.
*/

void
RealSubtractVectors(dest_vector, src_vector1, src_vector2, n)
double *dest_vector;
double *src_vector1;
double *src_vector2;
int n;

{
    int col;
    double *save_src1;
    double *save_src2;

    save_src1 = (double *) RealCreateVector(n);
    save_src2 = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src1, src_vector1, n);
    (void) RealCopyVector(save_src2, src_vector2, n);
    for(col = 0; col < n; col++) {
	dest_vector[col] = save_src1[col] - save_src2[col];
    }
    (void) RealFreeVector(save_src1);
    (void) RealFreeVector(save_src2);
}

/*
This function computes the inner product of two real-valued vectors.
*/

void
RealInnerProduct(res, src_vector1, src_vector2, n)
double *res;
double *src_vector1;
double *src_vector2;
int n;

{
    int index;
    double temp;

    temp = 0.0;
    for(index = 0; index < n; index++) {
	temp += src_vector1[index] * src_vector2[index];
    }
    *res = temp;
}

/*
This function computes the sum of all the elements of the given real-valued
vector.
*/

void
RealAccumulateVector(res, vector, n)
double *res;
double *vector;
int n;

{
    int index;
    double temp;

    temp = 0.0;
    for(index = 0; index < n; index++) {
	temp += vector[index];
    }
    *res = temp;
}

/*
This function performs the heap sort in the ascending order of all the elements
of the given real-valued vector.
*/

void
RealHeapSortVector(dest_vector, src_vector, n)
double *dest_vector;
double *src_vector;
int n;

{
    double *save_src;
    int index;

    save_src = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_src, src_vector, n);
    (void) BuildHeap(save_src, n);
    for(index = n - 1; index > 0; index--) {
	(void) SwapFloat(&(save_src[0]), &(save_src[index]));
	(void) Heapify(save_src, 0, index);
    }
    (void) RealCopyVector(dest_vector, save_src, n);
    (void) RealFreeVector(save_src);
}

/*
This function builds the heap for the use in the heap sort algorithm.
*/

void
BuildHeap(vector, n)
double *vector;
int n;

{
    int index;

    for(index = n / 2 - 1; index >= 0; index--) {
	(void) Heapify(vector, index, n);
    }
}

/*
This function organizes a vector into a heap for the use in heap sort.
*/

void
Heapify(vector, node, n)
double *vector;
int node;
int n;

{
    int left;
    int right;
    int largest;

    right = 2 * (node + 1);
    left = right - 1;
    if((left < n) && (vector[left] > vector[node])) {
	largest = left;
    } else {
	largest = node;
    }
    if((right < n) && (vector[right] > vector[largest])) {
	largest = right;
    }
    if(largest != node) {
	(void) SwapFloat(&(vector[node]), &(vector[largest]));
	(void) Heapify(vector, largest, n);
    }
}

/*
This function initializes all elements of the given [N by M] real-valued matrix to zero.
*/

void
RealClearMatrix(a, n, m)
double **a;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealClearVector(a[row], m);
    }
}

/*
This function allocates space for a new [N by M] real-valued matrix, 
initializes all of its elements to zero, and returns it on output.
*/

double **
RealCreateMatrix(n, m)
int n;
int m;

{
    double **a;
    int row;

    a = (double **) SafeMalloc((unsigned) (n) * sizeof(double *),
    "a", "RealCreateMatrix");
    for(row = 0; row < n; row++) {
	a[row] = (double *) RealCreateVector(m);
    }
    return(a);
}

/*
This function frees memory occupied bye a real-valued matrix.
*/

void
RealFreeMatrix(a, n)
double **a;
int n;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealFreeVector(a[row]);
    }
    (void) free(a);
}

/*
This function copies the contents of the source real-valued matrix into 
the destination real-valued matrix.  All matrices are [N by M].
*/

void
RealCopyMatrix(dest_matrix, src_matrix, n, m)
double **dest_matrix;
double **src_matrix;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealCopyVector(dest_matrix[row], src_matrix[row], m);
    }
}

/*
This function scales the contents of the source real-valued matrix into the 
destination real-valued matrix.  All matrices are [N by M].
*/

void
RealScaleMatrix(dest_matrix, src_matrix, scale, n, m)
double **dest_matrix;
double **src_matrix;
double scale;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealScaleVector(dest_matrix[row], src_matrix[row], scale, m);
    }
}

/*
This function initializes the contents of the destination real-valued matrix to
the real-valued scalar constant.
*/

void
RealInitMatrix(dest_matrix, value, n, m)
double **dest_matrix;
double value;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealInitVector(dest_matrix[row], value, m);
    }
}

/*
This function adds the contents of the two source real-valued matrices into 
the destination real-valued matrix.  All matrices are [N by M].
*/

void
RealAddMatrices(dest_matrix, src_matrix1, src_matrix2, n, m)
double **dest_matrix;
double **src_matrix1;
double **src_matrix2;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealAddVectors(dest_matrix[row], 
	src_matrix1[row], src_matrix2[row], m);
    }
}

/*
This function subtracts the contents of the second source real-valued matrix 
from the contents of the first source real-valued matrix into the destination 
real-valued matrix.  All matrices are [N by M].
*/

void
RealSubtractMatrices(dest_matrix, src_matrix1, src_matrix2, n, m)
double **dest_matrix;
double **src_matrix1;
double **src_matrix2;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) RealSubtractVectors(dest_matrix[row], 
	src_matrix1[row], src_matrix2[row], m);
    }
}

/*
This function multiplies the [N by M] real-valued matrix A by the [M by 1] 
real-valued vector x (supplied on input as a [1 by M] row vector) and puts the
result into the [N by 1] real-valued vector y (supplied on output as a [1 by N]
row vector).
*/

void
RealMatrixMultVector(y, a, x, n, m)
double *y;
double **a;
double *x;
int n;
int m;

{
    int row;
    double **save_a;
    double *save_x;
    
    save_a = (double **) RealCreateMatrix(n, m);
    save_x = (double *) RealCreateVector(m);
    (void) RealCopyMatrix(save_a, a, n, m);
    (void) RealCopyVector(save_x, x, m);
    for(row = 0; row < n; row++) {
	(void) RealInnerProduct(&(y[row]), save_a[row], save_x, m);
    }
    (void) RealFreeMatrix(save_a, n);
    (void) RealFreeVector(save_x);
}

/*
This function transposes the [N by M] real-valued matrix A into the [M by N] 
real-valued matrix B.
*/

void 
RealTranspose(b, a, n, m)
double **b;
double **a;
int n;
int m;

{
    int row;
    int col;
    double **save_a;

    save_a = (double **) RealCreateMatrix(n, m);
    (void) RealCopyMatrix(save_a, a, n, m);
    for(row = 0; row < m; row++) {
	for(col = 0; col < n; col++) {
	    b[row][col] = save_a[col][row];
	}
    }
    (void) RealFreeMatrix(save_a, n);
}

/*
This function computes the trace of the [N by N] real-valued matrix A.
*/

void 
RealTrace(res, a, n)
double *res;
double **a;
int n;

{
    int index;
    double temp;

    temp = 0.0;
    for(index = 0; index < n; index++) {
	temp += a[index][index];
    }
    *res = temp;
}

/*
This function computes the sum of all the elements of the given [N by M]
real-valued matrix A.
*/

void
RealAccumulateMatrix(res, a, n, m)
double *res;
double **a;
int n;
int m;

{
    int row;
    double temp1;
    double temp2;

    temp1 = 0.0;
    for(row = 0; row < n; row++) {
	(void) RealAccumulateVector(&temp2, a[row], m);
	temp1 += temp2;
    }
    *res = temp1;
}

/*
This function computes the inner product of two [N by M] real-valued 
matrices A and B.
*/

void
RealMatrixInnerProduct(res, a, b, n, m)
double *res;
double **a;
double **b;
int n;
int m;

{
    int row;
    double temp1;
    double temp2;

    temp1 = 0.0;
    for(row = 0; row < n; row++) {
	(void) RealInnerProduct(&temp2, a[row], b[row], m);
	temp1 += temp2;
    }
    *res = temp1;
}

/*
This function multiplies the [N by M] real-valued matrix A by the [M by L] 
real-valued matrix B and puts the result into the [N by L] real-valued matrix C.
*/

void
RealMatrixMultMatrix(c, a, b, n, m, l)
double **c;
double **a;
double **b;
int n;
int m;
int l;

{
    int row;
    double **temp_b_xpose;
    double **temp_c_xpose;
    
    temp_b_xpose = (double **) RealCreateMatrix(l, m);
    temp_c_xpose = (double **) RealCreateMatrix(l, n);
    (void) RealTranspose(temp_b_xpose, b, m, l);
    for(row = 0; row < l; row++) {
	(void) RealMatrixMultVector(temp_c_xpose[row], 
	a, temp_b_xpose[row], n, m);
    }
    (void) RealTranspose(c, temp_c_xpose, l, n);
    (void) RealFreeMatrix(temp_b_xpose, l);
    (void) RealFreeMatrix(temp_c_xpose, l);
}

/*
This function solves the real-valued matrix equation Ax = b (where A is a 
square [N by N] real-valued matrix) via the Gaussian elimination with partial 
pivoting.  First, the "forward elimination" converts A into an 
upper-triangular real-valued matrix.  Then, the "backward substitution" 
computes real-valued vector x.
*/

void
RealGaussElimPartPiv(x, a, b, n)
double *x;
double **a;
double *b;
int n;

{
    int index;
    int row;
    int col;
    double multiplier;
    double **save_a;
    double *save_b;

    save_a = (double **) RealCreateMatrix(n, n);
    save_b = (double *) RealCreateVector(n);
    (void) RealCopyMatrix(save_a, a, n, n);
    (void) RealCopyVector(save_b, b, n);
    for(index = 0; index < n; index++) {
	if((double) AbsFloat(save_a[index][index]) < EPSILON) {
	    (void) RealPartialPivot(save_a, save_b, index, n);
	}
	for(row = index + 1; row < n; row++) {
	    if(save_a[row][index] != 0.0) {
		multiplier = save_a[row][index] / save_a[index][index];
		for(col = index + 1; col < n; col++) {
		    save_a[row][col] -= multiplier * save_a[index][col];
		}
		save_b[row] -= multiplier * save_b[index];
	    }
	}
    }
    x[n-1] = save_b[n-1] / save_a[n-1][n-1];
    for(row = n - 2; row >= 0; row--) {
	for(col = row + 1, x[row] = 0.0; col < n; col++) {
	    x[row] -= save_a[row][col] * x[col];
	}
	x[row] = (x[row] + save_b[row]) / save_a[row][row];
    }
    (void) RealFreeMatrix(save_a, n);
    (void) RealFreeVector(save_b);
}

/*
This function assumes that the diagonal coefficient at location index is zero.
It finds the first row of A (where A is a square [N by N] real-valued matrix), 
in which the diagonal coefficient at location index is non-zero, and swaps the 
data in that row with the data in row index.  
*/

void 
RealPartialPivot(a, b, index, n)
double **a;
double *b;
int index;
int n;

{
    int row;
    int singular = 1;
    double *save_ptr;

    for(row = index + 1; (singular && (row < n)); row++) {
	if((double) AbsFloat(a[row][index]) > EPSILON) {
	    singular = 0;
	    save_ptr = a[index];
	    a[index] = a[row];
	    a[row] = save_ptr;
	    (void) SwapFloat(&(b[index]), &(b[row]));
	}
    }
    if(singular) {
	(void) fprintf(stdout, "Error:  Singular real matrix.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
}

/*
This function computes the inverse of the square [N by N] real-valued matrix A 
into the square [N by N] real-valued matrix B.
*/

void
RealInvertMatrix(b, a, n)
double **b;
double **a;
int n;

{
    double **x;
    double **i;
    int index;

    x = (double **) RealCreateMatrix(n, n);
    i = (double **) RealCreateMatrix(n, n);
    for(index = 0; index < n; index++) {
	i[index][index] = 1.0;
    }
    for(index = 0; index < n; index++) {
	(void) RealGaussElimPartPiv(x[index], a, i[index], n);
    }
    (void) RealTranspose(b, x, n, n);
    (void) RealFreeMatrix(x, n);
    (void) RealFreeMatrix(i, n);
}

/*
This function decomposes the given [N by M] real-valued matrix A into the 
product of two matrices: B [N by M] and C [M by M] by using the Gram-Schmidt 
process on the real-valued column vectors of A.  Matrix B has orthogonal
(or orthonormal) real-valued column vectors.  Matrix C is upper triangular 
with ones (or magnitudes) on the main diagonal.
*/

void
RealGramSchmidt(b, c, a, n, m, normalize)
double **b;
double **c;
double **a;
int n;
int m;
int normalize;

{
    int row;
    double **temp_a_xpose;

    temp_a_xpose = (double **) RealCreateMatrix(m, n);
    (void) RealTranspose(temp_a_xpose, a, n, m);
    (void) RealClearMatrix(c, m, m);
    if(normalize) {
	for(row = 0; row < m; row++) {
	    (void) RealOrthoNormVector(temp_a_xpose[row], 
	    temp_a_xpose[row], c[row], temp_a_xpose, row, n);
	}
    } else {
	for(row = 0; row < m; row++) {
	    (void) RealOrthogVector(temp_a_xpose[row], 
	    temp_a_xpose[row], c[row], temp_a_xpose, row, n);
	    c[row][row] = 1.0;
	}
    }
    (void) RealTranspose(b, temp_a_xpose, m, n);
    (void) RealTranspose(c, c, m, m);
    (void) RealFreeMatrix(temp_a_xpose, m);
}

/*
This function orthogonalizes the given [1 by M] real-valued row vector 
with respect to the basis B of N [1 by M] real-valued row vectors.
It also returns a vector that specifies the reversal of this process.
*/

void 
RealOrthogVector(dest_vector, src_vector, reverse, basis, n, m)
double *dest_vector;
double *src_vector;
double *reverse;
double **basis;
int n;
int m;

{
    int row;
    double projection;
    double normalizer;
    double *temp_vector;

    temp_vector = (double *) RealCreateVector(m);
    (void) RealCopyVector(dest_vector, src_vector, m);
    for(row = 0; row < n; row++) {
	(void) RealInnerProduct(&projection, src_vector, basis[row], m);
	(void) RealInnerProduct(&normalizer, basis[row], basis[row], m);
	projection /= (normalizer + EPSILON);
	reverse[row] = projection;
	(void) RealScaleVector(temp_vector, basis[row], projection, m);
	(void) RealSubtractVectors(dest_vector, dest_vector, temp_vector, m);
    }
    (void) RealFreeVector(temp_vector);
}

/*
This function orthonormalizes the given [1 by M] real-valued row vector 
with respect to the basis B of N [1 by M] real-valued row vectors.
It also returns a vector that specifies the reversal of this process.
*/

void 
RealOrthoNormVector(dest_vector, src_vector, reverse, basis, n, m)
double *dest_vector;
double *src_vector;
double *reverse;
double **basis;
int n;
int m;

{
    int row;
    double projection;
    double normalizer;
    double *temp_vector;

    temp_vector = (double *) RealCreateVector(m);
    (void) RealCopyVector(dest_vector, src_vector, m);
    for(row = 0; row < n; row++) {
	(void) RealInnerProduct(&projection, src_vector, basis[row], m);
	reverse[row] = projection;
	(void) RealScaleVector(temp_vector, basis[row], projection, m);
	(void) RealSubtractVectors(dest_vector, dest_vector, temp_vector, m);
    }
    (void) RealInnerProduct(&normalizer, dest_vector, dest_vector, m);
    normalizer = (double) sqrt(normalizer);
    projection = 1.0 / (normalizer + EPSILON);
    (void) RealScaleVector(dest_vector, dest_vector, projection, m);
    reverse[n] = normalizer;
    (void) RealFreeVector(temp_vector);
}

/*
This function computes the Eigenvector-Eigenvalue Decomposition of the real
symmetric non-negative-definite [N by N] matrix A: A = Q * D * Q'.  
Q is [N by N] and orthonormal; D is [N by N] and diagonal with the eigenvalues 
of A on the main diagonal.  Finally, D = Q' * A * Q.
*/

void
Eigen(q, d, a, n)
double **q;
double **d;
double **a;
int n;

{
    double **u;

    u = (double **) RealCreateMatrix(n, n);
    (void) SVD(u, d, q, a, n, n);
    (void) RealFreeMatrix(u, n);
}

/*
This function computes the Singular Value Decomposition of the [N by M]
matrix A: A = U * D * V'.  U is [N by M] and orthonormal; D is [M by M] and 
diagonal with positive entries; V is [M by M] and orthonormal.  Matrix V 
consists of the eigenvectors of A' * A.  The values in D are the square roots 
of the eigenvalues of A' * A.  Finally, U = A * D-inverse * V.
*/

void
SVD(u, d, v, a, n, m)
double **u;
double **d;
double **v;
double **a;
int n;
int m;

{
    double *svals;
    int index;

    svals = (double *) RealCreateVector(m);
    (void) RealCopyMatrix(u, a, n, m);
    (void) svdcmp_nric(u, n, m, svals, v);
    for(index = 0; index < m; index++) {
	d[index][index] = svals[index];
    }
    (void) RealFreeVector(svals);
}

/*
This is the actual singular value decomposition computational procedure.
It was taken from "Numerical Recipes In C" and modified to use the proper "C"
subscripts (beginning at 0).  It comes with two of its own macros.
*/

static double at, bt, ct; 
#define PYTHAG(a, b) \
((at = (double) AbsFloat(a)) > (bt = (double) AbsFloat(b)) ? \
(ct = bt / at, at * (double) sqrt(1.0 + ct * ct)) : \
(bt ? (ct = at / bt, bt * (double) sqrt(1.0 + ct * ct)): 0.0))
#define SIGN(a, b) ((b) >= 0.0 ? (double) AbsFloat(a) : -(double) AbsFloat(a))
void 
svdcmp_nric(a, m, n, w, v)
double **a; 
double *w; 
double **v; 
int m; 
int n; 

{
    int flag; 
    int i; 
    int its; 
    int j; 
    int jj; 
    int k; 
    int l; 
    int nm; 
    double c; 
    double f; 
    double h; 
    double s; 
    double x; 
    double y; 
    double z; 
    double anorm = 0.0; 
    double g = 0.0; 
    double scale = 0.0; 
    double *rv1; 

    if(m < n) {
	(void) fprintf(stdout,  "Error: Underdetermined matrix.  Exit 1.\n"); 
	(void) fflush(stdout); 
	(void) exit(1); 
    }
    rv1 = (double *) RealCreateVector(n); 
    for(i = 0; i < n; i++) {
        l = i + 1; 
        rv1[i] = scale * g; 
        g = s = scale = 0.0; 
        if(i < m) {
            for(k = i; k < m; k++) {
		scale += (double) AbsFloat(a[k][i]); 
	    }
            if(scale > EPSILON) {
                for(k = i; k < m; k++) {
                    a[k][i] /= scale; 
                    s += a[k][i] * a[k][i]; 
                }
                f = a[i][i]; 
                g = -SIGN((double) sqrt(s), f); 
                h = f * g - s; 
                a[i][i] = f - g; 
                if(i != (n - 1)) {
                    for(j = l; j < n; j++) {
                        for(s = 0.0, k = i; k < m; k++) {
			    s += a[k][i] * a[k][j]; 
			}
                        f = s / h; 
                        for(k = i; k < m; k++) {
			    a[k][j] += f * a[k][i]; 
			}
                    }
                }
                for(k = i; k < m; k++) {
		    a[k][i] *= scale; 
		}
            }
        }
        w[i] = scale * g; 
        g = s = scale = 0.0; 
        if((i < m) && (i != (n - 1))) {
            for(k = l; k < n; k++) {
		scale += (double) AbsFloat(a[i][k]); 
	    }
            if(scale > EPSILON) {
                for(k = l; k < n; k++) {
                    a[i][k] /= scale; 
                    s += a[i][k] * a[i][k]; 
                }
                f = a[i][l]; 
                g = -SIGN((double) sqrt(s), f); 
                h = f * g - s; 
                a[i][l] = f - g; 
                for(k = l; k < n; k++) {
		    rv1[k] = a[i][k] / h; 
		}
                if(i != (m - 1)) {
                    for(j = l; j < m; j++) {
                        for(s = 0.0, k = l; k < n; k++) {
			    s += a[j][k] * a[i][k]; 
			}
                        for(k = l; k < n; k++) {
			    a[j][k] += s * rv1[k]; 
			}
                    }
                }
                for(k = l; k < n; k++) {
		    a[i][k] *= scale; 
		}
            }
        }
        anorm = (double) MaxFloat(anorm, 
	((double) AbsFloat(w[i]) + (double) AbsFloat(rv1[i]))); 
    }
    for(i = n - 1; i >= 0; i--) {
        if(i < (n - 1)) {
            if((double) AbsFloat(g) > EPSILON) {
                for(j = l; j < n; j++) {
                    v[j][i] = (a[i][j] / a[i][l]) / g; 
		}
                for(j = l; j < n; j++) {
                    for(s = 0.0, k = l; k < n; k++) {
			s += a[i][k] * v[k][j]; 
		    }
                    for(k = l; k < n; k++) {
			v[k][j] += s * v[k][i]; 
		    }
                }
            }
            for(j = l; j < n; j++) {
		v[i][j] = v[j][i] = 0.0; 
	    }
        }
        v[i][i] = 1.0; 
        g = rv1[i]; 
        l = i; 
    }
    for(i = n - 1; i >= 0; i--) {
        l = i + 1; 
        g = w[i]; 
        if(i < (n - 1)) {
            for(j = l; j < n; j++) {
		a[i][j] = 0.0; 
	    }
	}
        if((double) AbsFloat(g) > EPSILON) {
            g = 1.0 / g; 
            if(i != (n - 1)) {
                for(j = l; j < n; j++) {
                    for(s = 0.0, k = l; k < m; k++) {
			s += a[k][i] * a[k][j]; 
		    }
                    f = (s / a[i][i]) * g; 
                    for(k = i; k < m; k++) {
			a[k][j] += f * a[k][i]; 
		    }
                }
            }
            for(j = i; j < m; j++) {
		a[j][i] *= g; 
	    }
        } else {
            for(j = i; j < m; j++) {
		a[j][i] = 0.0; 
	    }
        }
        ++(a[i][i]); 
    }
    for(k = n - 1; k >= 0; k--) {
        for(its = 0; its < 30; its++) {
            flag = 1; 
            for(l = k; l >= 0; l--) {
                nm = l - 1; 
                if(((double) AbsFloat(rv1[l]) + anorm) == anorm) {
                    flag = 0; 
                    break; 
                }
                if(((double) AbsFloat(w[nm]) + anorm) == anorm) {
		    break; 
		}
            }
            if(flag) {
                c = 0.0; 
                s = 1.0; 
                for(i = l; i <= k; i++) {
                    f = s * rv1[i]; 
                    if(((double) AbsFloat(f) + anorm) != anorm) {
                        g = w[i]; 
                        h = PYTHAG(f, g); 
                        w[i] = h; 
                        h = 1.0 / h; 
                        c = g * h; 
                        s = (-f * h); 
                        for(j = 0; j < m; j++) {
                            y = a[j][nm]; 
                            z = a[j][i]; 
                            a[j][nm] = y * c + z * s; 
                            a[j][i] = z * c - y * s; 
                        }
                    }
                }
            }
            z = w[k]; 
            if(l == k) {
                if(z < 0.0) {
                    w[k] = -z; 
                    for(j = 0; j < n; j++) {
			v[j][k] = (-v[j][k]); 
		    }
                }
                break; 
            }
            if(its == 30) {
		(void) fprintf(stdout,  
		"Error: No convergence in 30 SVDCMP iterations.  Exit 1.\n"); 
		(void) fflush(stdout); 
		(void) exit(1); 
	    }
            x = w[l]; 
            nm = k - 1; 
            y = w[nm]; 
            g = rv1[nm]; 
            h = rv1[k]; 
            f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y); 
            g = PYTHAG(f, 1.0); 
            f = ((x - z) * (x + z) + h * ((y / (f + SIGN(g, f))) - h)) / x; 
            c = s = 1.0; 
            for(j = l; j <= nm; j++) {
                i = j + 1; 
                g = rv1[i]; 
                y = w[i]; 
                h = s * g; 
                g = c * g; 
                z = PYTHAG(f, h); 
                rv1[j] = z; 
                c = f / z; 
                s = h / z; 
                f = x * c + g * s; 
                g = g * c - x * s; 
                h = y * s; 
                y = y * c; 
                for(jj = 0; jj < n; jj++) {
                    x = v[jj][j]; 
                    z = v[jj][i]; 
                    v[jj][j] = x * c + z * s; 
                    v[jj][i] = z * c - x * s; 
                }
                z = PYTHAG(f, h); 
                w[j] = z; 
                if((double) AbsFloat(z) > EPSILON) {
                    z = 1.0 / z; 
                    c = f * z; 
                    s = h * z; 
                }
                f = (c * g) + (s * y); 
                x = (c * y) - (s * g); 
                for(jj = 0; jj < m; jj++) {
                    y = a[jj][j]; 
                    z = a[jj][i]; 
                    a[jj][j] = y * c + z * s; 
                    a[jj][i] = z * c - y * s; 
                }
            }
            rv1[l] = 0.0; 
            rv1[k] = f; 
            w[k] = x; 
        }
    }
    (void) RealFreeVector(rv1); 
}
#undef SIGN
#undef PYTHAG

/*
This function recursively computes the determinant of the given [N by N]
(square) real-valued matrix.
*/

void
RealDeterminant(res, a, n)
double *res;
double **a;
int n;

{
    double **b;
    double minor_res;
    int m;
    int row;
    int sub_row;
    int sub_row_count;
    double temp;

    if(n > 1) {
	temp = 0.0;
	m = n - 1;
	b = (double **) RealCreateMatrix(m, m);
	for(row = 0; row < n; row++) {
	    if((double) AbsFloat(a[row][0]) > EPSILON) {
		for(sub_row = sub_row_count = 0; sub_row < n; sub_row++) {
		    if(sub_row != row) {
			(void) RealCopyVector(b[sub_row_count++],
			&(a[sub_row][1]), m);
		    }
		}
		(void) RealDeterminant(&minor_res, b, m);
		temp +=
		a[row][0] * ((int) Odd(row) ? (-minor_res) : minor_res);
	    }
	}
	(void) RealFreeMatrix(b, m);
    } else if(n == 1) {
	temp = a[0][0];
    } else if(n == 0) {
	temp = 0.0;
    } else {
	(void) fprintf(stdout, "Error: Illegal matrix size %d.  Exit 1.\n", n);
	(void) fflush(stdout);
	(void) exit(1);
    }
    *res = temp;
}

/*
This function prints the given [1 by N] real-valued row vector for the 
purposes of diagnostics.
*/

void
RealShowVector(name, v, n)
char *name;
double *v;
int n;

{
    int col;

    (void) fprintf(stdout, "Vector '%s':\n", name);
    (void) fflush(stdout);
    for(col = 0; col < n; col++) {
	(void) fprintf(stdout, "%g ", v[col]);
	(void) fflush(stdout);
    }
    (void) fprintf(stdout, "\n");
    (void) fflush(stdout);
}

/*
This function prints the given [N by M] real-valued matrix for the purposes of 
diagnostics.
*/

void
RealShowMatrix(name, a, n, m)
char *name;
double **a;
int n;
int m;

{
    int row;
    int col;

    (void) fprintf(stdout, "Matrix '%s':\n", name);
    (void) fflush(stdout);
    for(row = 0; row < n; row++) {
	for(col = 0; col < m; col++) {
	    (void) fprintf(stdout, "%g ", a[row][col]);
	    (void) fflush(stdout);
	}
	(void) fprintf(stdout, "\n");
	(void) fflush(stdout);
    }
}

/*
This function initializes all elements of the given complex-valued vector 
to zero.
*/

void
ComplexClearVector(v, n)
COMPLEX **v;
int n;

{
    int col;

    for(col = 0; col < n; col++) {
	(void) ComplexSet(v[col], 0.0, 0.0);
    }
}

/*
This function allocates space for a new complex-valued vector, initializes 
all of its elements to zero, and returns it on output.
*/

COMPLEX **
ComplexCreateVector(n)
int n;

{
    COMPLEX **v;
    int col;

    v = (COMPLEX **) SafeMalloc((unsigned) (n) * sizeof(COMPLEX *),
    "v", "ComplexCreateVector");
    for(col = 0; col < n; col++) {
	v[col] = (COMPLEX *) ComplexBuild(0.0, 0.0);
    }
    return(v);
}

/*
This function frees memory occupied bye a complex-valued vector.
*/

void
ComplexFreeVector(v, n)
COMPLEX **v;
int n;

{
    int index;

    for(index = 0; index < n; index++) {
	(void) ComplexFree(v[index]);
    }
    (void) free(v);
}

/*
This function copies the contents of the source complex-valued vector into the 
destination complex-valued vector.
*/

void
ComplexCopyVector(dest_vector, src_vector, n)
COMPLEX **dest_vector;
COMPLEX **src_vector;
int n;

{
    int col;
    COMPLEX **save_src;

    save_src = (COMPLEX **) ComplexCreateVector(n);
    for(col = 0; col < n; col++) {
	(void) ComplexCopy(save_src[col], src_vector[col]);
    }
    for(col = 0; col < n; col++) {
	(void) ComplexCopy(dest_vector[col], save_src[col]);
    }
    (void) ComplexFreeVector(save_src, n);
}

/*
This function scales by a constant complex number the contents of the source 
complex-valued vector into the destination complex-valued vector.
*/

void
ComplexScaleVector(dest_vector, src_vector, scale, n)
COMPLEX **dest_vector;
COMPLEX **src_vector;
COMPLEX *scale;
int n;

{
    int col;
    COMPLEX **save_src;

    save_src = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexCopyVector(save_src, src_vector, n);
    for(col = 0; col < n; col++) {
	(void) ComplexMultiply(dest_vector[col], save_src[col], scale);
    }
    (void) ComplexFreeVector(save_src, n);
}

/*
This function initializes the contents of the destination complex-valued vector
to the complex-valued scalar constant.
*/

void
ComplexInitVector(dest_vector, value, n)
COMPLEX **dest_vector;
COMPLEX *value;
int n;

{
    int col;

    for(col = 0; col < n; col++) {
	(void) ComplexCopy(dest_vector[col], value);
    }
}

/*
This function adds the contents of the two source complex-valued vectors into 
the destination complex-valued vector.
*/

void
ComplexAddVectors(dest_vector, src_vector1, src_vector2, n)
COMPLEX **dest_vector;
COMPLEX **src_vector1;
COMPLEX **src_vector2;
int n;

{
    int col;
    COMPLEX **save_src1;
    COMPLEX **save_src2;

    save_src1 = (COMPLEX **) ComplexCreateVector(n);
    save_src2 = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexCopyVector(save_src1, src_vector1, n);
    (void) ComplexCopyVector(save_src2, src_vector2, n);
    for(col = 0; col < n; col++) {
	(void) ComplexAdd(dest_vector[col], save_src1[col], save_src2[col]);
    }
    (void) ComplexFreeVector(save_src1, n);
    (void) ComplexFreeVector(save_src2, n);
}

/*
This function subtracts the contents of the second source complex-valued vector
from the contents of the first source complex-valued vector into the 
destination complex-valued vector.
*/

void
ComplexSubtractVectors(dest_vector, src_vector1, src_vector2, n)
COMPLEX **dest_vector;
COMPLEX **src_vector1;
COMPLEX **src_vector2;
int n;

{
    int col;
    COMPLEX **save_src1;
    COMPLEX **save_src2;

    save_src1 = (COMPLEX **) ComplexCreateVector(n);
    save_src2 = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexCopyVector(save_src1, src_vector1, n);
    (void) ComplexCopyVector(save_src2, src_vector2, n);
    for(col = 0; col < n; col++) {
	(void) ComplexSubtract(dest_vector[col], 
	save_src1[col], save_src2[col]);
    }
    (void) ComplexFreeVector(save_src1, n);
    (void) ComplexFreeVector(save_src2, n);
}

/*
This function complex conjugates the contents of the source complex-valued 
vector into the destination complex-valued vector.
*/

void
ComplexConjugateVector(dest_vector, src_vector, n)
COMPLEX **dest_vector;
COMPLEX **src_vector;
int n;

{
    int col;
    COMPLEX **save_src;

    save_src = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexCopyVector(save_src, src_vector, n);
    for(col = 0; col < n; col++) {
	(void) ComplexConjugate(dest_vector[col], save_src[col]);
    }
    (void) ComplexFreeVector(save_src, n);
}

/*
This function computes the inner product of two complex-valued vectors.
*/

void
ComplexInnerProduct(res, src_vector1, src_vector2, n)
COMPLEX *res;
COMPLEX **src_vector1;
COMPLEX **src_vector2;
int n;

{
    int index;
    COMPLEX *temp1;
    COMPLEX *temp2;

    temp1 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp2 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(index = 0; index < n; index++) {
	(void) ComplexMultiply(temp2, src_vector1[index], src_vector2[index]);
	(void) ComplexAdd(temp1, temp1, temp2);
    }
    (void) ComplexCopy(res, temp1);
    (void) ComplexFree(temp1);
    (void) ComplexFree(temp2);
}

/*
This function computes the complex conjugate inner product of two 
complex-valued vectors.
*/

void
ComplexConjugateInnerProduct(res, src_vector1, src_vector2, n)
COMPLEX *res;
COMPLEX **src_vector1;
COMPLEX **src_vector2;
int n;

{
    int index;
    COMPLEX *temp1;
    COMPLEX *temp2;

    temp1 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp2 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(index = 0; index < n; index++) {
	(void) ComplexConjugate(temp2, src_vector1[index]);
	(void) ComplexMultiply(temp2, temp2, src_vector2[index]);
	(void) ComplexAdd(temp1, temp1, temp2);
    }
    (void) ComplexCopy(res, temp1);
    (void) ComplexFree(temp1);
    (void) ComplexFree(temp2);
}

/*
This function computes the sum of all the elements of the given complex-valued
vector.
*/

void 
ComplexAccumulateVector(res, vector, n)
COMPLEX *res;
COMPLEX **vector;
int n;

{
    int index;
    COMPLEX *temp;

    temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(index = 0; index < n; index++) {
	(void) ComplexAdd(temp, temp, vector[index]);
    }
    (void) ComplexCopy(res, temp);
    (void) ComplexFree(temp);
}

/*
This function initializes all elements of the given [N by M] complex-valued 
matrix to zero.
*/

void
ComplexClearMatrix(a, n, m)
COMPLEX ***a;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexClearVector(a[row], m);
    }
}

/*
This function allocates space for a new [N by M] complex-valued matrix, 
initializes all of its elements to zero, and returns it on output.
*/

COMPLEX ***
ComplexCreateMatrix(n, m)
int n;
int m;

{
    COMPLEX ***a;
    int row;

    a = (COMPLEX ***) SafeMalloc((unsigned) (n) * sizeof(COMPLEX **),
    "a", "ComplexCreateMatrix");
    for(row = 0; row < n; row++) {
	a[row] = (COMPLEX **) ComplexCreateVector(m);
    }
    return(a);
}

/*
This function frees memory occupied bye a complex-valued matrix.
*/

void
ComplexFreeMatrix(a, n, m)
COMPLEX ***a;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexFreeVector(a[row], m);
    }
    (void) free(a);
}

/*
This function copies the contents of the source complex-valued matrix into 
the destination complex-valued matrix.  All matrices are [N by M].
*/

void
ComplexCopyMatrix(dest_matrix, src_matrix, n, m)
COMPLEX ***dest_matrix;
COMPLEX ***src_matrix;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexCopyVector(dest_matrix[row], src_matrix[row], m);
    }
}

/*
This function scales the contents of the source complex-valued matrix into the 
destination complex-valued matrix.  All matrices are [N by M].
*/

void
ComplexScaleMatrix(dest_matrix, src_matrix, scale, n, m)
COMPLEX ***dest_matrix;
COMPLEX ***src_matrix;
COMPLEX *scale;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexScaleVector(dest_matrix[row], src_matrix[row], scale, m);
    }
}

/*
This function initializes the contents of the destination complex-valued matrix
to the complex-valued scalar constant.
*/

void
ComplexInitMatrix(dest_matrix, value, n, m)
COMPLEX ***dest_matrix;
COMPLEX *value;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexInitVector(dest_matrix[row], value, m);
    }
}

/*
This function adds the contents of the two source complex-valued matrices into 
the destination complex-valued matrix.  All matrices are [N by M].
*/

void
ComplexAddMatrices(dest_matrix, src_matrix1, src_matrix2, n, m)
COMPLEX ***dest_matrix;
COMPLEX ***src_matrix1;
COMPLEX ***src_matrix2;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexAddVectors(dest_matrix[row], 
	src_matrix1[row], src_matrix2[row], m);
    }
}

/*
This function subtracts the contents of the second source complex-valued 
matrix from the contents of the first source complex-valued matrix into the 
destination complex-valued matrix.  All matrices are [N by M].
*/

void
ComplexSubtractMatrices(dest_matrix, src_matrix1, src_matrix2, n, m)
COMPLEX ***dest_matrix;
COMPLEX ***src_matrix1;
COMPLEX ***src_matrix2;
int n;
int m;

{
    int row;

    for(row = 0; row < n; row++) {
	(void) ComplexSubtractVectors(dest_matrix[row], 
	src_matrix1[row], src_matrix2[row], m);
    }
}

/*
This function complex conjugates the contents of the source complex-valued 
matrix into the destination complex-valued matrix.  All matrices are [N by M].
*/

void
ComplexConjugateMatrix(dest_matrix, src_matrix, n, m)
COMPLEX ***dest_matrix;
COMPLEX ***src_matrix;
int n;
int m;

{
    int row;
    COMPLEX ***save_src;

    save_src = (COMPLEX ***) ComplexCreateMatrix(n, m);
    (void) ComplexCopyMatrix(save_src, src_matrix, n, m);
    for(row = 0; row < n; row++) {
	(void) ComplexConjugateVector(dest_matrix[row], save_src[row], m);
    }
    (void) ComplexFreeMatrix(save_src, n, m);
}

/*
This function multiplies the [N by M] complex-valued matrix A by the [M by 1] 
complex-valued vector x (supplied on input as a [1 by M] row vector) and puts 
the result into the [N by 1] complex-valued vector y (supplied on output as 
a [1 by N] row vector).
*/

void
ComplexMatrixMultVector(y, a, x, n, m)
COMPLEX **y;
COMPLEX ***a;
COMPLEX **x;
int n;
int m;

{
    int row;
    COMPLEX ***save_a;
    COMPLEX **save_x;
    
    save_a = (COMPLEX ***) ComplexCreateMatrix(n, m);
    save_x = (COMPLEX **) ComplexCreateVector(m);
    (void) ComplexCopyMatrix(save_a, a, n, m);
    (void) ComplexCopyVector(save_x, x, m);
    for(row = 0; row < n; row++) {
	(void) ComplexInnerProduct(y[row], save_a[row], save_x, m);
    }
    (void) ComplexFreeMatrix(save_a, n, m);
    (void) ComplexFreeVector(save_x, m);
}

/*
This function transposes the [N by M] complex-valued matrix A into the [M by N] complex-valued matrix B.
*/

void 
ComplexTranspose(b, a, n, m)
COMPLEX ***b;
COMPLEX ***a;
int n;
int m;

{
    int row;
    int col;
    COMPLEX ***save_a;

    save_a = (COMPLEX ***) ComplexCreateMatrix(n, m);
    (void) ComplexCopyMatrix(save_a, a, n, m);
    for(row = 0; row < m; row++) {
	for(col = 0; col < n; col++) {
	    (void) ComplexCopy(b[row][col], save_a[col][row]);
	}
    }
    (void) ComplexFreeMatrix(save_a, n, m);
}

/*
This function complex conjugates and transposes the [N by M] complex-valued 
matrix A into the [M by N] complex-valued matrix B.
*/

void 
ComplexConjugateTranspose(b, a, n, m)
COMPLEX ***b;
COMPLEX ***a;
int n;
int m;

{
    int row;
    int col;
    COMPLEX ***save_a;

    save_a = (COMPLEX ***) ComplexCreateMatrix(n, m);
    (void) ComplexCopyMatrix(save_a, a, n, m);
    (void) ComplexConjugateMatrix(save_a, save_a, n, m);
    for(row = 0; row < m; row++) {
	for(col = 0; col < n; col++) {
	    (void) ComplexCopy(b[row][col], save_a[col][row]);
	}
    }
    (void) ComplexFreeMatrix(save_a, n, m);
}

/*
This function computes the trace of the [N by N] complex-valued matrix A.
*/

void
ComplexTrace(res, a, n)
COMPLEX *res;
COMPLEX ***a;
int n;

{
    int index;
    COMPLEX *temp;

    temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(index = 0; index < n; index++) {
	(void) ComplexAdd(temp, temp, a[index][index]);
    }
    (void) ComplexCopy(res, temp);
    (void) ComplexFree(temp);
}

/*
This function computes the sum of all the elements of the given [N by M]
complex-valued matrix A.
*/

void 
ComplexAccumulateMatrix(res, a, n, m)
COMPLEX *res;
COMPLEX ***a;
int n;
int m;

{
    int row;
    COMPLEX *temp1;
    COMPLEX *temp2;

    temp1 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp2 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(row = 0; row < n; row++) {
	(void) ComplexAccumulateVector(temp2, a[row], m);
	(void) ComplexAdd(temp1, temp1, temp2);
    }
    (void) ComplexCopy(res, temp1);
    (void) ComplexFree(temp1);
    (void) ComplexFree(temp2);
}

/*
This function computes the inner product of two [N by M] complex-valued 
matrices A and B.
*/

void
ComplexMatrixInnerProduct(res, a, b, n, m)
COMPLEX *res;
COMPLEX ***a;
COMPLEX ***b;
int n;
int m;

{
    int row;
    COMPLEX *temp1;
    COMPLEX *temp2;

    temp1 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp2 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    for(row = 0; row < n; row++) {
	(void) ComplexInnerProduct(temp2, a[row], b[row], m);
	(void) ComplexAdd(temp1, temp1, temp2);
    }
    (void) ComplexCopy(res, temp1);
    (void) ComplexFree(temp1);
    (void) ComplexFree(temp2);
}

/*
This function multiplies the [N by M] complex-valued matrix A by the [M by L] 
complex-valued matrix B and puts the result into the [N by L] complex-valued 
matrix C.
*/

void
ComplexMatrixMultMatrix(c, a, b, n, m, l)
COMPLEX ***c;
COMPLEX ***a;
COMPLEX ***b;
int n;
int m;
int l;

{
    int row;
    COMPLEX ***temp_b_xpose;
    COMPLEX ***temp_c_xpose;
    
    temp_b_xpose = (COMPLEX ***) ComplexCreateMatrix(l, m);
    temp_c_xpose = (COMPLEX ***) ComplexCreateMatrix(l, n);
    (void) ComplexTranspose(temp_b_xpose, b, m, l);
    for(row = 0; row < l; row++) {
	(void) ComplexMatrixMultVector(temp_c_xpose[row], 
	a, temp_b_xpose[row], n, m);
    }
    (void) ComplexTranspose(c, temp_c_xpose, l, n);
    (void) ComplexFreeMatrix(temp_b_xpose, l, m);
    (void) ComplexFreeMatrix(temp_c_xpose, l, n);
}

/*
This function solves the complex-valued matrix equation Ax = b (where A is a
square [N by N] complex-valued matrix) via the Gaussian elimination with 
partial pivoting.  First, the "forward elimination" converts A into an 
upper-triangular complex-valued matrix.  Then, the "backward substitution" 
computes complex-valued vector x.
*/

void
ComplexGaussElimPartPiv(x, a, b, n)
COMPLEX **x;
COMPLEX ***a;
COMPLEX **b;
int n;

{
    int index;
    int row;
    int col;
    COMPLEX *multiplier;
    COMPLEX ***save_a;
    COMPLEX **save_b;
    COMPLEX *temp;

    multiplier = (COMPLEX *) ComplexBuild(0.0, 0.0);
    save_a = (COMPLEX ***) ComplexCreateMatrix(n, n);
    save_b = (COMPLEX **) ComplexCreateVector(n);
    temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    (void) ComplexCopyMatrix(save_a, a, n, n);
    (void) ComplexCopyVector(save_b, b, n);
    for(index = 0; index < n; index++) {
	if((double) ComplexMagnitude(save_a[index][index]) < EPSILON) {
	    (void) ComplexPartialPivot(save_a, save_b, index, n);
	}
	for(row = index + 1; row < n; row++) {
	    if((double) ComplexMagnitude(save_a[row][index]) != 0.0) {
		(void) ComplexDivide(multiplier, 
		save_a[row][index], save_a[index][index]);
		for(col = index + 1; col < n; col++) {
		    (void) ComplexMultiply(temp, 
		    multiplier, save_a[index][col]);
		    (void) ComplexSubtract(save_a[row][col], 
		    save_a[row][col], temp);
		}
		(void) ComplexMultiply(temp, multiplier, save_b[index]);
		(void) ComplexSubtract(save_b[row], save_b[row], temp);
	    }
	}
    }
    (void) ComplexDivide(x[n-1], save_b[n-1], save_a[n-1][n-1]);
    for(row = n - 2; row >= 0; row--) {
	(void) ComplexSet(x[row], 0.0, 0.0);
	for(col = row + 1; col < n; col++) {
	    (void) ComplexMultiply(temp, save_a[row][col], x[col]);
	    (void) ComplexSubtract(x[row], x[row], temp);
	}
	(void) ComplexAdd(temp, x[row], save_b[row]);
	(void) ComplexDivide(x[row], temp, save_a[row][row]);
    }
    (void) ComplexFree(multiplier);
    (void) ComplexFreeMatrix(save_a, n, n);
    (void) ComplexFreeVector(save_b, n);
    (void) ComplexFree(temp);
}

/*
This function assumes that the diagonal coefficient at location index is zero.
It finds the first row of A (where A is a square [N by N] complex-valued 
matrix), in which the diagonal coefficient at location index is non-zero, and 
swaps the data in that row with the data in row index.  
*/

void 
ComplexPartialPivot(a, b, index, n)
COMPLEX ***a;
COMPLEX **b;
int index;
int n;

{
    int row;
    int singular = 1;
    COMPLEX **save_ptr;

    for(row = index + 1; (singular && (row < n)); row++) {
	if((double) ComplexMagnitude(a[row][index]) > EPSILON) {
	    singular = 0;
	    save_ptr = a[index];
	    a[index] = a[row];
	    a[row] = save_ptr;
	    (void) ComplexSwap(b[index], b[row]);
	}
    }
    if(singular) {
	(void) fprintf(stdout, "Error:  Singular complex matrix.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
}

/*
This function computes the inverse of the square [N by N] complex-valued 
matrix A into the square [N by N] complex-valued matrix B.
*/

void
ComplexInvertMatrix(b, a, n)
COMPLEX ***b;
COMPLEX ***a;
int n;

{
    COMPLEX ***x;
    COMPLEX ***i;
    int index;

    x = (COMPLEX ***) ComplexCreateMatrix(n, n);
    i = (COMPLEX ***) ComplexCreateMatrix(n, n);
    for(index = 0; index < n; index++) {
	(void) ComplexSet(i[index][index], 1.0, 0.0);
    }
    for(index = 0; index < n; index++) {
	(void) ComplexGaussElimPartPiv(x[index], a, i[index], n);
    }
    (void) ComplexTranspose(b, x, n, n);
    (void) ComplexFreeMatrix(x, n, n);
    (void) ComplexFreeMatrix(i, n, n);
}

/*
This function decomposes the given [N by M] complex-valued matrix A into the 
product of two matrices: B [N by M] and C [M by M] by using the Gram-Schmidt 
process on the complex-valued column vectors of A.  Matrix B has orthogonal
(or orthonormal) complex-valued column vectors.  Matrix C is upper triangular 
with ones (or magnitudes) on the main diagonal.
*/

void
ComplexGramSchmidt(b, c, a, n, m, normalize)
COMPLEX ***b;
COMPLEX ***c;
COMPLEX ***a;
int n;
int m;
int normalize;

{
    int row;
    COMPLEX ***temp_a_xpose;

    temp_a_xpose = (COMPLEX ***) ComplexCreateMatrix(m, n);
    (void) ComplexTranspose(temp_a_xpose, a, n, m);
    (void) ComplexClearMatrix(c, m, m);
    if(normalize) {
	for(row = 0; row < m; row++) {
	    (void) ComplexOrthoNormVector(temp_a_xpose[row], 
	    temp_a_xpose[row], c[row], temp_a_xpose, row, n);
	}
    } else {
	for(row = 0; row < m; row++) {
	    (void) ComplexOrthogVector(temp_a_xpose[row], 
	    temp_a_xpose[row], c[row], temp_a_xpose, row, n);
	    (void) ComplexSet(c[row][row], 1.0, 0.0);
	}
    }
    (void) ComplexTranspose(b, temp_a_xpose, m, n);
    (void) ComplexTranspose(c, c, m, m);
    (void) ComplexFreeMatrix(temp_a_xpose, m, n);
}

/*
This function orthogonalizes the given [1 by M] complex-valued row vector 
with respect to the basis B of N [1 by M] complex-valued row vectors.
It also returns a vector that specifies the reversal of this process.
*/

void 
ComplexOrthogVector(dest_vector, src_vector, reverse, basis, n, m)
COMPLEX **dest_vector;
COMPLEX **src_vector;
COMPLEX **reverse;
COMPLEX ***basis;
int n;
int m;

{
    int row;
    COMPLEX *projection;
    COMPLEX *normalizer;
    COMPLEX **temp_vector;

    projection = (COMPLEX *) ComplexBuild(0.0, 0.0);
    normalizer = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp_vector = (COMPLEX **) ComplexCreateVector(m);
    (void) ComplexCopyVector(dest_vector, src_vector, m);
    for(row = 0; row < n; row++) {
	(void) ComplexConjugateInnerProduct(projection, 
	src_vector, basis[row], m);
	(void) ComplexConjugateInnerProduct(normalizer, 
	basis[row], basis[row], m);
	(void) ComplexDivide(projection, projection, normalizer);
	(void) ComplexCopy(reverse[row], projection);
	(void) ComplexScaleVector(temp_vector, basis[row], projection, m);
	(void) ComplexSubtractVectors(dest_vector, 
	dest_vector, temp_vector, m);
    }
    (void) ComplexFree(projection);
    (void) ComplexFree(normalizer);
    (void) ComplexFreeVector(temp_vector, m);
}

/*
This function orthonormalizes the given [1 by M] complex-valued row vector 
with respect to the basis B of N [1 by M] complex-valued row vectors.
It also returns a vector that specifies the reversal of this process.
*/

void 
ComplexOrthoNormVector(dest_vector, src_vector, reverse, basis, n, m)
COMPLEX **dest_vector;
COMPLEX **src_vector;
COMPLEX **reverse;
COMPLEX ***basis;
int n;
int m;

{
    int row;
    COMPLEX *projection;
    COMPLEX *normalizer;
    COMPLEX **temp_vector;

    projection = (COMPLEX *) ComplexBuild(0.0, 0.0);
    normalizer = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp_vector = (COMPLEX **) ComplexCreateVector(m);
    (void) ComplexCopyVector(dest_vector, src_vector, m);
    for(row = 0; row < n; row++) {
	(void) ComplexConjugateInnerProduct(projection, 
	src_vector, basis[row], m);
	(void) ComplexCopy(reverse[row], projection);
	(void) ComplexScaleVector(temp_vector, basis[row], projection, m);
	(void) ComplexSubtractVectors(dest_vector, 
	dest_vector, temp_vector, m);
    }
    (void) ComplexConjugateInnerProduct(normalizer, 
    dest_vector, dest_vector, m);
    normalizer->real_part = (double) sqrt(normalizer->real_part);
    (void) ComplexSet(projection, 1.0, 0.0);
    (void) ComplexScale(projection, projection, 
    1.0 / (normalizer->real_part + EPSILON));
    (void) ComplexScaleVector(dest_vector, dest_vector, projection, m);
    (void) ComplexCopy(reverse[n], normalizer);
    (void) ComplexFree(projection);
    (void) ComplexFree(normalizer);
    (void) ComplexFreeVector(temp_vector, m);
}

/*
This function recursively computes the determinant of the given [N by N]
(square) complex-valued matrix.
*/

void
ComplexDeterminant(res, a, n)
COMPLEX *res;
COMPLEX ***a;
int n;

{
    COMPLEX ***b;
    COMPLEX *minor_res;
    int m;
    int row;
    int sub_row;
    int sub_row_count;
    COMPLEX *temp;

    temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    if(n > 1) {
	minor_res = (COMPLEX *) ComplexBuild(0.0, 0.0);
	m = n - 1;
	b = (COMPLEX ***) ComplexCreateMatrix(m, m);
	for(row = 0; row < n; row++) {
	    if((double) ComplexMagnitude(a[row][0]) > EPSILON) {
		for(sub_row = sub_row_count = 0; sub_row < n; sub_row++) {
		    if(sub_row != row) {
			(void) ComplexCopyVector(b[sub_row_count++],
			&(a[sub_row][1]), m);
		    }
		}
		(void) ComplexDeterminant(minor_res, b, m);
		(void) ComplexMultiply(minor_res, minor_res, a[row][0]);
		if((int) Odd(row)) {
		    (void) ComplexSubtract(temp, temp, minor_res);
		} else {
		    (void) ComplexAdd(temp, temp, minor_res);
		}
	    }
	}
	(void) ComplexFree(minor_res);
	(void) ComplexFreeMatrix(b, m, m);
    } else if(n == 1) {
	(void) ComplexCopy(temp, a[0][0]);
    } else if(n < 0) {
	(void) fprintf(stdout, "Error: Illegal matrix size %d.  Exit 1.\n", n);
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) ComplexCopy(res, temp);
    (void) ComplexFree(temp);
}

/*
This function prints the given [1 by N] complex-valued row vector for the 
purposes of diagnostics.
*/

void
ComplexShowVector(name, v, n)
char *name;
COMPLEX **v;
int n;

{
    int col;

    (void) fprintf(stdout, "Vector '%s':\n", name);
    (void) fflush(stdout);
    for(col = 0; col < n; col++) {
	(void) fprintf(stdout, "[%g,%g] ", 
	v[col]->real_part, v[col]->imaginary_part);
	(void) fflush(stdout);
    }
    (void) fprintf(stdout, "\n");
    (void) fflush(stdout);
}

/*
This function prints the given [N by M] complex-valued matrix for the purposes 
of diagnostics.
*/

void
ComplexShowMatrix(name, a, n, m)
char *name;
COMPLEX ***a;
int n;
int m;

{
    int row;
    int col;

    (void) fprintf(stdout, "Matrix '%s':\n", name);
    (void) fflush(stdout);
    for(row = 0; row < n; row++) {
	for(col = 0; col < m; col++) {
	    (void) fprintf(stdout, "[%g,%g] ", 
	    a[row][col]->real_part, a[row][col]->imaginary_part);
	    (void) fflush(stdout);
	}
	(void) fprintf(stdout, "\n");
	(void) fflush(stdout);
    }
}

/*
This function converts the given real data vector into the complex vector 
format.
*/

void
RealVectorToComplexVector(complex_vector, real_vector, n)
COMPLEX **complex_vector;
double *real_vector;
int n;

{
    int index;

    for(index = 0; index < n; index++) {
	(void) ComplexSet(complex_vector[index], real_vector[index], 0.0);
    }
}

/*
This function converts the given complex data vector into the real vector 
format.
*/

void
ComplexVectorToRealVector(real_vector, complex_vector, directive, n)
double *real_vector;
COMPLEX **complex_vector;
int directive;
int n;

{
    int index;

    if(directive == MAGNITUDE) {
	for(index = 0; index < n; index++) {
	    real_vector[index] = 
	    (double) ComplexMagnitude(complex_vector[index]);
	}
    } else if(directive == REALPART) {
	for(index = 0; index < n; index++) {
	    real_vector[index] = complex_vector[index]->real_part;
	}
    } else {
	(void) fprintf(stdout, 
	"Error: Illegal matrix conversion directive.  Exit 1.\n");
	(void) fflush(stdout);
	(void) exit(1);
    }
}

/*
This function converts the given real data matrix into the complex matrix 
format.
*/

void
RealMatrixToComplexMatrix(complex_matrix, real_matrix, height, width)
COMPLEX ***complex_matrix;
double **real_matrix;
int height;
int width;

{
    int row;

    for(row = 0; row < height; row++) {
	(void) RealVectorToComplexVector(complex_matrix[row], 
	real_matrix[row], width);
    }
}

/*
This function converts the given complex data matrix into the real matrix 
format.
*/

void
ComplexMatrixToRealMatrix(real_matrix, complex_matrix, directive,
height, width)
double **real_matrix;
COMPLEX ***complex_matrix;
int directive;
int height;
int width;

{
    int row;

    for(row = 0; row < height; row++) {
	(void) ComplexVectorToRealVector(real_matrix[row], 
	complex_matrix[row], directive, width);
    }
}

/*
This function converts the given real data vector into the real matrix format.
*/

void
RealVectorToRealMatrix(real_matrix, real_vector, height, width)
double **real_matrix;
double *real_vector;
int height;
int width;

{
    int row;
    double *real_temp_vector;
    int n;

    n = height * width;
    real_temp_vector = (double *) RealCreateVector(n);
    (void) RealCopyVector(real_temp_vector, real_vector, n);
    for(row = 0; row < height; row++) {
	(void) RealCopyVector(real_matrix[row], 
	&(real_vector[row*width]), width);
    }
    (void) RealFreeVector(real_temp_vector);
}

/*
This function converts the given real data vector into the complex matrix 
format.
*/

void
RealVectorToComplexMatrix(complex_matrix, real_vector, height, width)
COMPLEX ***complex_matrix;
double *real_vector;
int height;
int width;

{
    double **real_temp_matrix;

    real_temp_matrix = (double **) RealCreateMatrix(height, width);
    (void) RealVectorToRealMatrix(real_temp_matrix,
    real_vector, height, width);
    (void) RealMatrixToComplexMatrix(complex_matrix,
    real_temp_matrix, height, width);
    (void) RealFreeMatrix(real_temp_matrix, height);
}

/*
This function converts the given real data matrix into the real vector format.
*/

void
RealMatrixToRealVector(real_vector, real_matrix, height, width)
double *real_vector;
double **real_matrix;
int height;
int width;

{
    int row;
    double **real_temp_matrix;

    real_temp_matrix = (double **) RealCreateMatrix(height, width);
    (void) RealCopyMatrix(real_temp_matrix, real_matrix, height, width);
    for(row = 0; row < height; row++) {
	(void) RealCopyVector(&(real_vector[row*width]), 
	real_temp_matrix[row], width);
    }
    (void) RealFreeMatrix(real_temp_matrix, height);
}

/*
This function converts the given real data matrix into the complex vector 
format.
*/

void
RealMatrixToComplexVector(complex_vector, real_matrix, height, width)
COMPLEX **complex_vector;
double **real_matrix;
int height;
int width;

{
    double *real_temp_vector;
    int n;

    n = height * width;
    real_temp_vector = (double *) RealCreateVector(n);
    (void) RealMatrixToRealVector(real_temp_vector,
    real_matrix, height, width);
    (void) RealVectorToComplexVector(complex_vector, real_temp_vector, n);
    (void) RealFreeVector(real_temp_vector);
}

/*
This function converts the given complex data vector into the real matrix 
format.
*/

void
ComplexVectorToRealMatrix(real_matrix,
complex_vector, directive, height, width)
double **real_matrix;
COMPLEX **complex_vector;
int directive;
int height;
int width;

{
    double *real_temp_vector;
    int n;

    n = height * width;
    real_temp_vector = (double *) RealCreateVector(n);
    (void) ComplexVectorToRealVector(real_temp_vector,
    complex_vector, directive, n);
    (void) RealVectorToRealMatrix(real_matrix,
    real_temp_vector, height, width);
    (void) RealFreeVector(real_temp_vector);
}

/*
This function converts the given complex data vector into the complex matrix 
format.
*/

void
ComplexVectorToComplexMatrix(complex_matrix, complex_vector, height, width)
COMPLEX ***complex_matrix;
COMPLEX **complex_vector;
int height;
int width;

{
    int row;
    COMPLEX **complex_temp_vector;
    int n;

    n = height * width;
    complex_temp_vector = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexCopyVector(complex_temp_vector, complex_vector, n);
    for(row = 0; row < height; row++) {
	(void) ComplexCopyVector(complex_matrix[row], 
	&(complex_vector[row*width]), width);
    }
    (void) ComplexFreeVector(complex_temp_vector, n);
}

/*
This function converts the given complex data matrix into the real vector 
format.
*/

void
ComplexMatrixToRealVector(real_vector, complex_matrix, directive, height, width)
double *real_vector;
COMPLEX ***complex_matrix;
int directive;
int height;
int width;

{
    COMPLEX **complex_temp_vector;
    int n;

    n = height * width;
    complex_temp_vector = (COMPLEX **) ComplexCreateVector(n);
    (void) ComplexMatrixToComplexVector(complex_temp_vector,
    complex_matrix, height, width);
    (void) ComplexVectorToRealVector(real_vector,
    complex_temp_vector, directive, n);
    (void) ComplexFreeVector(complex_temp_vector, n);
}

/*
This function converts the given complex data matrix into the complex vector 
format.
*/

void
ComplexMatrixToComplexVector(complex_vector, complex_matrix, height, width)
COMPLEX **complex_vector;
COMPLEX ***complex_matrix;
int height;
int width;

{
    int row;
    COMPLEX ***complex_temp_matrix;

    complex_temp_matrix = (COMPLEX ***) ComplexCreateMatrix(height, width);
    (void) ComplexCopyMatrix(complex_temp_matrix,
    complex_matrix, height, width);
    for(row = 0; row < height; row++) {
	(void) ComplexCopyVector(&(complex_vector[row*width]), 
	complex_temp_matrix[row], width);
    }
    (void) ComplexFreeMatrix(complex_temp_matrix, height, width);
}
