/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "ols.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function decomposes the given [N by M] real-valued matrix A into the 
product of two matrices: B [N by L] and C [L by L], where L <= M.  Matrix B
has orthogonal real-valued column vectors.  Matrix C is upper triangular with 
ones on the main diagonal.  Matrix A corresponds to approximating the 
real-valued [1 by N] data vector X using the LLSE method.  The column vectors 
of the matrix A are the values of M basis functions evaluated at N data 
points.  Thus the objective is to choose a subset of all basis functions, L,
and the matrices B and C optimally in order to guarantee the required 
approximation error tolerance E.  The Gram-Schmidt process on the real-valued 
column vectors of A is used for orthogonalization.  
*/

double
RealOLS(b, c, a, p, l, energy, x, e, n, m)
double **b;
double **c;
double **a;
int *p;/*parameter-index list*/
int *l;
double *energy;
double *x;/*data to be approximated*/
double e;
int n;
int m;

{
    int row;
    int *useless;
    double **temp_a_xpose;
    double **temp_b_xpose;
    double **b_xpose;
    double **temp_c_xpose;
    double gain;
    double max_gain;
    int best_row;
    double error;

    useless = (int *) SafeMalloc((unsigned) m * sizeof(int), 
    "useless", "RealOLS");
    temp_a_xpose = (double **) RealCreateMatrix(m, n);
    temp_b_xpose = (double **) RealCreateMatrix(m, n);
    b_xpose = (double **) RealCreateMatrix(m, n);
    temp_c_xpose = (double **) RealCreateMatrix(m, m);
    (void) RealTranspose(temp_a_xpose, a, n, m);
    (void) RealClearMatrix(c, m, m);
    (void) RealInnerProduct(energy, x, x, n);
    e *= *energy;
    best_row = (int) NIL;
    for(row = 0; row < m; row++) {
	c[row][row] = 1.0;
	p[row] = (int) NIL;
	useless[row] = 0;
    }
/*
printf("energy=%g;target-error=%g\n",*energy,e);fflush(stdout);
*/
    for(*l = 0; 
    (((error = (double) RealGlobalError(x, b_xpose, *l, *energy, n)) > e) && 
    (*l < m)); (*l)++) {
/*
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}
*/
	for(row = 0, max_gain = 0.0; row < m; row++) {
	    if(!useless[row]) {
		(void) RealOrthogVector(temp_b_xpose[row], temp_a_xpose[row],
		temp_c_xpose[row], b_xpose, *l, n);
		(void) RealInnerProduct(&gain, 
		temp_b_xpose[row], temp_b_xpose[row], n);
		if(gain < EPSILON) {
		    useless[row] = 1;
		} else {
		    gain = (double) RealFindGain(x, temp_b_xpose[row], 
		    *energy, n);
		    if(gain > max_gain) {
			max_gain = gain;
			best_row = row;
		    }
		}
	    }
	}
	if(best_row != (int) NIL) {
	    useless[best_row] = 1;
	    p[*l] = best_row;
	    (void) RealCopyVector(b_xpose[*l], temp_b_xpose[best_row], n);
	    (void) RealCopyVector(c[*l], temp_c_xpose[best_row], *l);
	}
    }
/*
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}else{
printf("after basis %d; error=%g\n",(*l),error);fflush(stdout);
}
*/
    (void) RealTranspose(b, b_xpose, *l, n);
    (void) RealTranspose(c, c, *l, *l);
    (void) free(useless);
    (void) RealFreeMatrix(temp_a_xpose, m);
    (void) RealFreeMatrix(temp_b_xpose, m);
    (void) RealFreeMatrix(b_xpose, m);
    (void) RealFreeMatrix(temp_c_xpose, m);
    return(error);
}

/*
This function decomposes the given [N by M] real-valued matrix A into the 
product of two matrices: B [N by L] and C [L by L], where L <= M.  Matrix B
has orthonormal real-valued column vectors.  Matrix C is upper triangular with 
magnitudes on the main diagonal.  Matrix A corresponds to approximating the 
real-valued [1 by N] data vector X using the LLSE method.  The column vectors 
of the matrix A are the values of M basis functions evaluated at N data 
points.  Thus the objective is to choose a subset of all basis functions, L,
and the matrices B and C optimally in order to guarantee the required 
approximation error tolerance E.  The Gram-Schmidt process on the real-valued 
column vectors of A is used for orthonormalization.  
*/

double
RealONLS(b, c, a, p, l, energy, x, e, n, m)
double **b;
double **c;
double **a;
int *p;/*parameter-index list*/
int *l;
double *energy;
double *x;/*data to be approximated*/
double e;
int n;
int m;

{
    int row;
    int *useless;
    double **temp_a_xpose;
    double **temp_b_xpose;
    double **b_xpose;
    double **temp_c_xpose;
    double gain;
    double max_gain;
    int best_row;
    double error;

    useless = (int *) SafeMalloc((unsigned) m * sizeof(int), 
    "useless", "RealOLS");
    temp_a_xpose = (double **) RealCreateMatrix(m, n);
    temp_b_xpose = (double **) RealCreateMatrix(m, n);
    b_xpose = (double **) RealCreateMatrix(m, n);
    temp_c_xpose = (double **) RealCreateMatrix(m, m);
    (void) RealTranspose(temp_a_xpose, a, n, m);
    (void) RealClearMatrix(c, m, m);
    (void) RealInnerProduct(energy, x, x, n);
    e *= *energy;
    best_row = (int) NIL;
    for(row = 0; row < m; row++) {
	p[row] = (int) NIL;
	useless[row] = 0;
    }
/*
printf("energy=%g;target-error=%g\n",*energy,e);fflush(stdout);
*/
    for(*l = 0; 
    (((error = (double) RealGlobalError(x, b_xpose, *l, *energy, n)) > e) && 
    (*l < m)); (*l)++) {
/*
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}
*/
	for(row = 0, max_gain = 0.0; row < m; row++) {
	    if(!useless[row]) {
		(void) RealOrthoNormVector(temp_b_xpose[row], 
		temp_a_xpose[row], temp_c_xpose[row], b_xpose, *l, n);
		gain = (double) RealFindGain(x, temp_b_xpose[row], *energy, n);
		if(gain > max_gain) {
		    max_gain = gain;
		    best_row = row;
		}
	    }
	}
	if(best_row != (int) NIL) {
	    useless[best_row] = 1;
	    p[*l] = best_row;
	    (void) RealCopyVector(b_xpose[*l], temp_b_xpose[best_row], n);
	    (void) RealCopyVector(c[*l], temp_c_xpose[best_row], *l+1);
	}
    }
/*
if(*l){
printf("after basis %d; error=%g\n",(*l)-1,error);fflush(stdout);
}else{
printf("after basis %d; error=%g\n",(*l),error);fflush(stdout);
}
*/
    (void) RealTranspose(b, b_xpose, *l, n);
    (void) RealTranspose(c, c, *l, *l);
    (void) free(useless);
    (void) RealFreeMatrix(temp_a_xpose, m);
    (void) RealFreeMatrix(temp_b_xpose, m);
    (void) RealFreeMatrix(b_xpose, m);
    (void) RealFreeMatrix(temp_c_xpose, m);
    return(error);
}

/*
This function computes the squared error in approximating a [1 by N]
real-valued data vector with L real-valued [1 by N] basis vectors.
*/

double
RealGlobalError(x, b, l, energy, n)
double *x;
double **b;
int l;
double energy;
int n;

{
    int index;
    double error;

    for(index = 0, error = energy; index < l; index++) {
	error -= (double) RealFindGain(x, b[index], energy, n);
    }
    return((double) AbsFloat(error));
}

/*
This function computes the squared gain in approximating a [1 by N]
real-valued data vector with a [1 by N] real-valued basis vector.
*/

double
RealFindGain(x, b, energy, n)
double *x;
double *b;
double energy;
int n;

{
    double temp1;
    double temp2;

    if(energy < EPSILON) {
	return(energy);
    }
    (void) RealInnerProduct(&temp1, b, x, n);
    (void) RealInnerProduct(&temp2, b, b, n);
    return((temp1 * temp1) / (temp2 + EPSILON));
}

/*
This function decomposes the given [N by M] complex-valued matrix A into the 
product of two matrices: B [N by L] and C [L by L], where L <= M.  Matrix B
has orthonormal complex-valued column vectors.  Matrix C is upper triangular 
with magnitudes on the main diagonal.  Matrix A corresponds to approximating 
the complex-valued [1 by N] data vector X using the LLSE method.  The column 
vectors of the matrix A are the values of M basis functions evaluated at N 
data points.  Thus the objective is to choose a subset of all basis functions, 
L, and the matrices B and C optimally in order to guarantee the required 
approximation error tolerance E.  The Gram-Schmidt process on the 
complex-valued column vectors of A is used for orthonormalization. 
*/

double
ComplexONLS(b, c, a, p, l, real_energy, x, e, n, m)
COMPLEX ***b;
COMPLEX ***c;
COMPLEX ***a;
int *p;/*parameter-index list*/
int *l;
double *real_energy;
COMPLEX **x;/*data to be approximated*/
double e;
int n;
int m;

{
    int row;
    int *useless;
    COMPLEX ***temp_a_xpose;
    COMPLEX ***temp_b_xpose;
    COMPLEX ***b_xpose;
    COMPLEX ***temp_c_xpose;
    COMPLEX *temp;
    double gain;
    double max_gain;
    COMPLEX *energy;
    int best_row;
    double error;

    useless = (int *) SafeMalloc((unsigned) m * sizeof(int), 
    "useless", "ComplexOLS");
    temp_a_xpose = (COMPLEX ***) ComplexCreateMatrix(m, n);
    temp_b_xpose = (COMPLEX ***) ComplexCreateMatrix(m, n);
    b_xpose = (COMPLEX ***) ComplexCreateMatrix(m, n);
    temp_c_xpose = (COMPLEX ***) ComplexCreateMatrix(m, m);
    temp = (COMPLEX *) ComplexBuild(0.0, 0.0);
    energy = (COMPLEX *) ComplexBuild(0.0, 0.0);
    (void) ComplexConjugateInnerProduct(energy, x, x, n);
    *real_energy = energy->real_part;
    e *= *real_energy;
    (void) ComplexTranspose(temp_a_xpose, a, n, m);
    (void) ComplexClearMatrix(c, m, m);
    best_row = (int) NIL;
    for(row = 0; row < m; row++) {
	p[row] = (int) NIL;
	useless[row] = 0;
    }
    for(*l = 0; 
    (((error = (double) ComplexGlobalError(x, b_xpose, *l, energy, n)) > e) && 
    (*l < m)); (*l)++) {
	for(row = 0, max_gain = 0.0; row < m; row++) {
	    if(!useless[row]) {
		(void) ComplexOrthoNormVector(temp_b_xpose[row], 
		temp_a_xpose[row], temp_c_xpose[row], b_xpose, *l, n);
		gain = (double) ComplexFindGain(x, temp_b_xpose[row], 
		energy, n);
		if(gain > max_gain) {
		    max_gain = gain;
		    best_row = row;
		}
	    }
	}
	if(best_row != (int) NIL) {
	    useless[best_row] = 1;
	    p[*l] = best_row;
	    (void) ComplexCopyVector(b_xpose[*l], temp_b_xpose[best_row], n);
	    (void) ComplexCopyVector(c[*l], temp_c_xpose[best_row], *l+1);
	}
    }
    (void) ComplexTranspose(b, b_xpose, *l, n);
    (void) ComplexTranspose(c, c, *l, *l);
    (void) free(useless);
    (void) ComplexFreeMatrix(temp_a_xpose, m, n);
    (void) ComplexFreeMatrix(temp_b_xpose, m, n);
    (void) ComplexFreeMatrix(b_xpose, m, n);
    (void) ComplexFreeMatrix(temp_c_xpose, m, m);
    (void) ComplexFree(temp);
    (void) ComplexFree(energy);
    return(error);
}

/*
This function computes the squared error in approximating a [1 by N]
complex-valued data vector with L complex-valued [1 by N] basis vectors.
*/

double
ComplexGlobalError(x, b, l, energy, n)
COMPLEX **x;
COMPLEX ***b;
int l;
COMPLEX *energy;
int n;

{
    int index;
    double error;

    for(index = 0, error = energy->real_part; index < l; index++) {
	error -= (double) ComplexFindGain(x, b[index], energy, n);
    }
    return(error);
}

/*
This function computes the squared gain in approximating a [1 by N]
complex-valued vector with a [1 by N] complex-valued basis vector.
*/

double
ComplexFindGain(x, b, energy, n)
COMPLEX **x;
COMPLEX **b;
COMPLEX *energy;
int n;

{
    COMPLEX *temp1;
    COMPLEX *temp2;
    COMPLEX *temp3;
    double gain;

    temp1 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp2 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    temp3 = (COMPLEX *) ComplexBuild(0.0, 0.0);
    if((double) ComplexMagnitude(energy) < EPSILON) {
	return(energy->real_part);
    }
    (void) ComplexConjugateInnerProduct(temp1, b, x, n);
    (void) ComplexConjugateInnerProduct(temp2, b, b, n);
    (void) ComplexDivide(temp1, temp1, temp2);
    (void) ComplexConjugate(temp3, temp1);
    (void) ComplexMultiply(temp1, temp1, temp3);
    (void) ComplexMultiply(temp1, temp1, temp2);
    gain = temp1->real_part;
    (void) ComplexFree(temp1);
    (void) ComplexFree(temp2);
    (void) ComplexFree(temp3);
    return(gain);
}
