/*
*******************************************************************************
Author: 
	Alex Sherstinsky,
	Media Laboratory, 
	Building E15 Room 383,
	20 Ames Street, Cambridge, MA 02139,
	Massachusetts Institute Of Technology,
	77 Massachusetts Avenue,
	Cambridge, MA 02139
	Telephone: (617)253-0381 
	Electronic Mail Address: shers@media.mit.edu
Copyright 1993 Massachusetts Institute Of Technology
All rights reserved.
This program may not be distributed to other organizations without the 
permission of the author.  In no case may it be distributed outside the
United States and Canada.  This program is supplied "as is", without any
kind of warranty, and M.I.T. does not promise to support it in any way.
This program is not to be incorporated into a commercial product, or access
to it sold, without prior written agreement from the author.
*******************************************************************************
AlMa Library
*******************************************************************************
*/
#include "stdutils.h"
#include "simplex.h"
#include "complex.h"
#include "matrix.h"
#include "func.h"

/*
This function implements the Downhill Simplex In Multidimensions algorithm.  
The [(D + 1) by D] matrix contains the (D + 1) starting vertices of the 
simplex in the [1 by D] parameter vector space.  The [1 by (D + 1)] vector 
contains the values of the objective function (to be minimized) at these 
vertices.  The optimal parameter vectors and the corresponding values of the 
objective function are produced on output.
*/

void
DownhillSimplex(parms, num_eval, objective_function, d, low, high, difftol)
double *parms;
int *num_eval;
double (*objective_function)();
int d;
double low;
double high;
double difftol;

{
    double **vertices;
    double *values;
    double *centroid_vertex;
    double *reflection_vertex;
    double *expansion_vertex;
    double *contraction_vertex;
    double centroid_value;
    double reflection_value;
    double expansion_value;
    double contraction_value;
    double error;
    int num_iter;
    int num_vertices;
    int low_index;
    int high_index;
    int next_high_index;
    int index;

    num_vertices = d + 1;
    vertices = (double **) RealCreateMatrix(num_vertices, d);
    values = (double *) RealCreateVector(num_vertices);
    centroid_vertex = (double *) RealCreateVector(d);
    reflection_vertex = (double *) RealCreateVector(d);
    expansion_vertex = (double *) RealCreateVector(d);
    contraction_vertex = (double *) RealCreateVector(d);
    (void) InitSimplex(vertices, values, num_eval, 
    objective_function, low, high, num_vertices, d);
    for(num_iter = 0, error = INFINITY;
    ((error > difftol) && (num_iter < MAXITER)); num_iter++) {
	(void) OrderSimplex(&low_index, &high_index, &next_high_index, 
	values, num_vertices);
	(void) RealCopyVector(parms, vertices[low_index], d);
	(void) FindCentroid(centroid_vertex, 
	vertices, high_index, num_vertices, d);
	centroid_value = 
	(double) (*objective_function)(centroid_vertex);
	(*num_eval)++;
	error = (double) FindError(values, centroid_value, num_vertices);
printf("ITER %d;error=%g\n",num_iter,error);fflush(stdout);
	(void) Extrapolate(reflection_vertex, vertices[high_index], 
	centroid_vertex, REFLECTCOEFF, d);
	reflection_value = 
	(double) (*objective_function)(reflection_vertex);
	(*num_eval)++;
	if(reflection_value < values[low_index]) {
	    (void) Extrapolate(expansion_vertex, reflection_vertex,
	    centroid_vertex, EXPANDCOEFF, d);
	    expansion_value = 
	    (double) (*objective_function)(expansion_vertex);
	    (*num_eval)++;
	    if(expansion_value < values[low_index]) {
		(void) RealCopyVector(vertices[high_index], 
		expansion_vertex, d);
		values[high_index] = expansion_value;
	    } else {
		(void) RealCopyVector(vertices[high_index], 
		reflection_vertex, d);
		values[high_index] = reflection_value;
	    }
	} else if(reflection_value > values[next_high_index]) {
	    if(reflection_value < values[high_index]) {
		(void) RealCopyVector(vertices[high_index], 
		reflection_vertex, d);
		values[high_index] = reflection_value;
	    }
	    (void) Extrapolate(contraction_vertex, vertices[high_index], 
	    centroid_vertex, CONTRACTCOEFF, d);
	    contraction_value = 
	    (double) (*objective_function)(contraction_vertex);
	    (*num_eval)++;
	    if(contraction_value > values[high_index]) {
		(void) ContractTowardLow(vertices, values, num_eval, 
		objective_function, low_index, num_vertices, d);
	    } else {
		(void) RealCopyVector(vertices[high_index], 
		contraction_vertex, d);
		values[high_index] = contraction_value;
	    }
	} else {
	    (void) RealCopyVector(vertices[high_index], 
	    reflection_vertex, d);
	    values[high_index] = reflection_value;
	}
    }
    if(num_iter == MAXITER) {
	(void) fprintf(stdout, "Error: No convergence in simplex.  Exit 1.\n"); 
	(void) fflush(stdout);
	(void) exit(1);
    }
    (void) RealFreeMatrix(vertices, num_vertices);
    (void) RealFreeVector(values);
    (void) RealFreeVector(centroid_vertex);
    (void) RealFreeVector(reflection_vertex);
    (void) RealFreeVector(expansion_vertex);
    (void) RealFreeVector(contraction_vertex);
}

/*
This function extrapolates vector1 through vector2 by the given factor.
All vectors are [1 by N] row vectors.
*/

void
Extrapolate(result, vector1, vector2, factor, n)
double *result;
double *vector1;
double *vector2;
double factor;
int n;

{
    double *save_vector1;
    double *save_vector2;

    save_vector1 = (double *) RealCreateVector(n);
    save_vector2 = (double *) RealCreateVector(n);
    (void) RealCopyVector(save_vector1, vector1, n);
    (void) RealCopyVector(save_vector2, vector2, n);
    (void) RealSubtractVectors(result, save_vector1, save_vector2, n);
    (void) RealScaleVector(result, result, factor, n);
    (void) RealAddVectors(result, result, save_vector2, n);
    (void) RealFreeVector(save_vector1);
    (void) RealFreeVector(save_vector2);
}

/*
This function contracts all vertices of the simplex toward its low point.
*/

void 
ContractTowardLow(vertices, values, num_eval, 
objective_function, low_index, num_vertices, d)
double **vertices;
double *values;
int *num_eval;
double (*objective_function)();
int low_index;
int num_vertices;
int d;

{
    int index;

    for(index = 0; index < num_vertices; index++) {
	if(index != low_index) {
	    (void) RealAddVectors(vertices[index], 
	    vertices[index], vertices[low_index], d);
	    (void) RealScaleVector(vertices[index], 
	    vertices[index], CONTRACTCOEFF, d);
	    values[index] = 
	    (double) (*objective_function)(vertices[index]);
	    (*num_eval)++;
	}
    }
}

/*
This function computes the centroid of (N - 1) [1 by M] row vectors.  
The cetroid is found by minimizing the cumulative Euclidian distance to it 
from the (N - 1) vectors.  The vectors are supplied in the [N by M] matrix.
*/

void
FindCentroid(centroid, matrix, exclude, n, m)
double *centroid;
double **matrix;
int exclude;
int n;
int m;

{
    int row;
    double **temp_matrix;

    temp_matrix = (double **) RealCreateMatrix(n, m);
    (void) RealCopyMatrix(temp_matrix, matrix, n, m);
    (void) RealClearVector(centroid, m);
    for(row = 0; row < n; row++) {
	if(row != exclude) {
	    (void) RealAddVectors(centroid, centroid, temp_matrix[row], m);
	}
    }
    (void) RealScaleVector(centroid, centroid, (1.0 / (n - 1)), m);
    (void) RealFreeMatrix(temp_matrix, n);
}

/*
This function orders the vertices of the simplex according to the values of 
the objective function at their positions.
*/

void
OrderSimplex(low_index, high_index, next_high_index, values, num_vertices)
int *low_index;
int *high_index;
int *next_high_index;
double *values;
int num_vertices;

{
    int index;

    for(index = 0, *low_index = *high_index = 0; 
    index < num_vertices; index++) {
	if(values[index] < values[*low_index]) {
	    *low_index = index;
	}
	if(values[index] > values[*high_index]) {
	    *high_index = index;
	}
    }
    for(index = 0, *next_high_index = *low_index; 
    index < num_vertices; index++) {
	if((index != *high_index) &&
	(values[index] > values[*next_high_index])) {
	    *next_high_index = index;
	}
    }
}

/*
This function determines the average difference between the values of the
objective function at the vertices of the simplex and its centroid.
*/

double
FindError(values, centroid_value, num_vertices)
double *values;
double centroid_value;
int num_vertices;

{
    int index;
    double difference;

    for(index = 0, difference = 0.0; 
    index < num_vertices; index++) {
	difference += (values[index] - centroid_value) * 
	(values[index] - centroid_value);
    }
    difference /= num_vertices;
    return(difference = (double) sqrt(difference));
}

/*
This function initializes the simplex geometrical figure.
The [(D + 1) by D] matrix contains the (D + 1) starting vertices of the 
simplex in the D-dimensional vector space. 
*/

void
InitSimplex(vertices, values, num_eval, 
objective_function, low, high, num_vertices, d)
double **vertices;
double *values;
int *num_eval;
double (*objective_function)();
double low;
double high;
int num_vertices;
int d;

{
    int index;
    int row;
    int col;

    (void) RealClearMatrix(vertices, num_vertices, d);
    /*
    for(index = 1; index < num_vertices; index++) {
	vertices[index][index - 1] = 1.0;
    }
    */
    for(row = 0; row < num_vertices; row++) {
	for(col = 0; col < d; col++) {
	    vertices[row][col] = low + (high - low) * (double) Random();
	}
    }
    for(index = 0; index < num_vertices; index++) {
	values[index] = 
	(double) (*objective_function)(vertices[index]);
	(*num_eval)++;
    }
}
